#include "dataController.h"

dataController::dataController()
{
	//Initialize the dataObject to NULL
	thisModel             = NULL;
	thisEmittanceAnalysis = NULL;
	thisMomentAnalysis    = NULL;
	master                = NULL;
};

dataController::~dataController()
{
	;
};

void dataController::setMaster(masterController* mc)
{
	master = mc;
}

void dataController::clearModel()
{
	if( isNotEmpty() )
	{
		delete thisModel;
		thisModel = NULL;
	}
}

bool dataController::isEmpty()
{
	return thisModel == NULL;
}

bool dataController::isNotEmpty()
{
	return thisModel != NULL;
}

bool dataController::fillFrom1DArray(float* start, profile_data_header_type* head)
{
	clearModel();
	thisModel = new dataObject( master->getCurrentMachine(), master->getCurrentTomography() );
	thisModel->bunchFromArray( start, head );

	return true;
}

bool dataController::readInPRPFile(const char* filename)
{
	bool didRead = false;
	clearModel(); //delete dataObject set thisModel = NULL;

	thisModel = new dataObject( master->getCurrentMachine(), master->getCurrentTomography() );

	//Need to initialize the model to know how much of the file to read in.
	//This would probably be better handled by reading in the whole thing and then having
	//dataObject parse it's values based on updated parameters.
	if( thisModel->readInPRPFile( filename ) )
	{
		didRead = true;
	}else{
		//Delete object and reset NULL pointer if reading failed.
		clearModel();
	}

	return didRead;
}


void dataController::tomography()
{
	char status[256];

	if( thisModel != NULL )
	{
		complete_data_header_type config = thisModel->getConfig();
	 	std::string mapLocation = TARDIS_Maps;
	 	master->statusUpdate("Locating or Creating maps...");

	 	MapClass maps( config, mapLocation, MapClass::kSparseSingleUse );

	 	tomography( &maps );
	}

}

void dataController::tomography( MapClass* maps )
{
	char status[256];

	if( thisModel != NULL )
	{
		complete_data_header_type config = thisModel->getConfig();
		int algo = config.tomography.algorithm;
		int progress = 0;

		thisModel->prepForTomography();

		#pragma omp parallel for shared(maps,progress)
		for(int bunch = 0; bunch < config.dataAcquisition.num_bunches; ++bunch)
		{
		 	thisModel->reconstructSingle( maps, bunch, algo );

		 	#pragma omp critical //It's likely that progress is causing a crash.
		 	{
		 		++progress;
			 	if(progress % 5 == 0 )
			 	{
			 		sprintf(status, "bunch %d/%d", progress, config.dataAcquisition.num_bunches);
			 		master->statusUpdate( status );
			 	}
		 	}
		}
		//end omp loop-----------------------

		sprintf(status, "bunch %d/%d", progress, config.dataAcquisition.num_bunches);

	}else{
	 	sprintf(status, "No data for this tab.");
	}

	 master->statusUpdate( status );
}

void dataController::createHistogram(const int bunch, TVirtualPad* tvPad, const int type)
{
	createRangeHistogram( bunch, 1, tvPad, type ); //Checks for NULL pointer.
}

void dataController::createRangeHistogram(const int bunch1, const int range, TVirtualPad* tvPad, const int type)
{

	tvPad->Clear();

	if(thisModel != NULL)
	{
		//Float_t norm  = 0;
		profile_data_header_type head = getDataAcquisition();
		Tomography_Config tomoHead    = getTomography();

		float* array;
		float BunchSpace = (float)(1e+9)*1/(getMachine()).RF_f;
		int xPts, yPts;
		float xMin, xMax, yMin, yMax;
		char axesLabels[256];
		switch(type){
			case kImage:
			{
				float bucketHeightMEV = (float)(1e-6)*BucketHeight(getMachine());
				xPts =  thisModel->getImageWidth();
				yPts =  thisModel->getImageHeight();
				xMin =  0.0;
				xMax =  BunchSpace*range;
				yMin = -bucketHeightMEV;
				yMax =  bucketHeightMEV;
				sprintf(axesLabels,"c1 Histo; ns ;#Delta E[MeV]");
			}
			break;
			case kProj:
			{
				xPts = head.ptsperprofile;
				yPts = (getTomography()).Frames;
				xMin = 0;
				xMax = BunchSpace*range;
				yMin = 0.0;
				yMax = (float)(getTomography()).Frames*head.turns_per_trig;
				sprintf(axesLabels,"c1 Histo; ns ;turns");
			}
			break;
			case kData:
			{
				xPts = head.ptsperprofile;
				yPts = (thisModel->getConfig()).tomography.Frames;
				xMin = 0;
				xMax = BunchSpace*range;
				yMin = 0.0;
				yMax = (float)(getTomography()).Frames*head.turns_per_trig;
				sprintf(axesLabels,"c1 Histo; ns ;turns");
			}
			break;
		}

		if(range == 1)
		{
			xMin -= xMax/2.0;
			xMax -= xMax/2.0;
		}

		Float_t value = 0;
		TH2F* hist = new TH2F("Image",(const char*)&axesLabels, range*xPts, xMin, xMax, yPts, yMin, yMax);
		for(int r = 0; r < range; ++r)
		{
			if(bunch1 + r < head.num_bunches)
			{
				int xStart = r * xPts;
				switch(type)
				{
					case kImage: array = thisModel->getImage( bunch1 + r );
					break;
					case kProj:  array = thisModel->getProfile( bunch1 + r );
					break;
					case kData:  array = thisModel->getData( bunch1 + r );
					break;
				}

				for(int y = 0; y < yPts; ++y)
				{
					int yStart = y*xPts;
					for(int x = 0; x < xPts; ++x)
					{
						value = array[ x+yStart ];
						hist->SetBinContent( x+1 + xStart, y+1, value );
					}
				}
			}
		}
		char title[100];
		//No more popping titles!
		hist->GetXaxis()->CenterTitle();
		hist->GetYaxis()->CenterTitle();
		hist->GetXaxis()->SetTitleSize(0.04);
		hist->GetYaxis()->SetTitleSize(0.04);

		if( range == 1 )
		{
			sprintf(title, "Bunches %d ", bunch1 + 1 );
			hist->SetTitle(title);

			hist->Draw("COLZ");
		}else{
			sprintf(title, "Bunches %d - %d", bunch1 + 1, (bunch1+1)+(range-1) );
			hist->SetTitle(title);
			hist->Draw("COL");
		}
	  	tvPad->Update();
	}
}

void dataController::fillHistogram(const int bunch, TH2F* hist, const int type)
{
	fillRangeHistogram( bunch, 1, hist, type); //Checks for NULL pointer.
}

void dataController::fillRangeHistogram(const int firstBunch, const int bunchRange, TH2F* hist, const int type)
{

	if( thisModel != NULL )
	{
		profile_data_header_type head = getDataAcquisition();
		Tomography_Config tomoHead  = getTomography();
		//This is all to handle firstBunch = -1 being interpreted as average bunch
		int plotRange    = firstBunch == -1 ? 1                : bunchRange;
		int bunchStart   = firstBunch == -1 ? 0                : firstBunch;
		int endBunchLoop = firstBunch == -1 ? head.num_bunches : bunchRange;

		//Float_t norm  = 0;

		float* array;
		float BunchSpace = (float)(1e+9)*1/(getMachine()).RF_f;
		int xPts, yPts;
		float xMin, xMax, yMin, yMax;
		std::string titleWhich;
		char axesLabels[256];
		switch(type){
			case kImage:
			{
				float bucketHeightMEV = (float)(1e-6)*BucketHeight(getMachine());
				xPts =  thisModel->getImageWidth();
				yPts =  thisModel->getImageHeight();
				xMin =  0.0;
				xMax =  BunchSpace*plotRange;
				yMin = -bucketHeightMEV;
				yMax =  bucketHeightMEV;
				titleWhich = "";
				sprintf(axesLabels,"c1 Histo; ns ;#Delta E[MeV]");
			}
			break;
			case kProj:
			{
				xPts = head.ptsperprofile;
				yPts = (getTomography()).Frames;
				xMin = 0;
				xMax = BunchSpace*plotRange;
				yMin = 0.0;
				yMax = (float)(getTomography()).Frames*head.turns_per_trig;
				titleWhich = " - Output";
				sprintf(axesLabels,"c1 Histo; ns ;turns");
			}
			break;
			case kData:
			{
				xPts = head.ptsperprofile;
				yPts = (thisModel->getConfig()).tomography.Frames;
				xMin = 0;
				xMax = BunchSpace*plotRange;
				yMin = 0.0;
				yMax = (float)(getTomography()).Frames*head.turns_per_trig;
				titleWhich = " - Input";
				sprintf(axesLabels,"c1 Histo; ns ;turns");
			}
			break;
		}

		if(plotRange == 1)
		{
			xMin -= xMax/2.0;
			xMax -= xMax/2.0;
		}

		Float_t value = 0;
		hist->Reset();
		hist->SetNameTitle("Image",(const char*)&axesLabels);
		hist->SetBins( plotRange*xPts, xMin, xMax, yPts, yMin, yMax );

		for(int r = 0; r < endBunchLoop; ++r)
		{
			if(bunchStart + r < head.num_bunches)
			{
				int xStart = firstBunch == -1 ? 0 : r * xPts;
				switch(type)
				{
					case kImage: array = thisModel->getImage( bunchStart + r );
					break;
					case kProj:  array = thisModel->getProfile( bunchStart + r );
					break;
					case kData:  array = thisModel->getData( bunchStart + r );
					break;
				}

				for(int y = 0; y < yPts; ++y)
				{
					int yStart = y * xPts;
					for(int x = 0; x < xPts; ++x)
					{
						//Adding to existing bin content lets you
						//calculate average bunch if specified.
						value = hist->GetBinContent( x+1 + xStart, y+1);
						value += array[ x+yStart ];
						hist->SetBinContent( x+1 + xStart, y+1, value );
					}
				}
			}
		}
		char title[100];
		//No more popping titles!
		hist->GetXaxis()->CenterTitle();
		hist->GetYaxis()->CenterTitle();
		hist->GetXaxis()->SetTitleSize(0.04);
		hist->GetYaxis()->SetTitleSize(0.04);

		if( firstBunch == -1 )
		{
			sprintf(title, "Average Bunch %s", titleWhich.c_str());
		}else if( plotRange == 1){
			sprintf(title, "Bunch %d %s", bunchStart + 1, titleWhich.c_str() );
		}else{
			sprintf(title, "Bunches %d - %d %s", bunchStart + 1, (bunchStart+1)+(endBunchLoop-1), titleWhich.c_str() );
		}
		hist->SetTitle(title);
	}
}

void dataController::fillEmittanceHisto(const int bunch, const int range, TH2F* hist)
{
	complete_data_header_type thisModelConfig = thisModel->getConfig();
	const int num_bunches = ( thisModel->getDataAcquisition() ).num_bunches;

	if(num_bunches != 0 )
	{
		const int emitPoints = 20;

		thisEmittanceAnalysis = new emittanceAnalysis( &thisModelConfig, thisModel );
		thisEmittanceAnalysis->calculateEmittanceContour(emitPoints);

		std::vector< double >* contourValues = thisEmittanceAnalysis->getContourValues();
		int numberPoints = contourValues->size();

		hist->Reset();
		hist->SetNameTitle("emittanceName","Emittance; bunch; #Delta E[MeV]");
		//numberPoints = 21; //Yikes! why is this hardcoded?
		if( numberPoints > 0 )
		{
			hist->SetBins( num_bunches, 0.0, (float)num_bunches, numberPoints, 0.0, (float)(*contourValues)[numberPoints-1]/(1E6) );

			for(int bunch = 0; bunch < num_bunches; ++bunch)
			{
				std::vector< double >* bunchEmittance = thisEmittanceAnalysis->getEmittanceCurves(bunch);

				for( int i=0; i < numberPoints; i++ ) hist->SetBinContent( bunch+1, i + 1, (float)(*bunchEmittance)[i] );
			}

			hist->GetXaxis()->CenterTitle();
			hist->GetYaxis()->CenterTitle();
			hist->GetXaxis()->SetDecimals();
			hist->GetYaxis()->SetDecimals();
			hist->GetXaxis()->SetNdivisions(30);
			hist->GetYaxis()->SetNdivisions(10);
			hist->GetXaxis()->SetTitleSize(0.04);
			hist->GetYaxis()->SetTitleSize(0.04);
		}

		delete thisEmittanceAnalysis;
	}
}

void dataController::fillEmittanceGraph(const int bunch, const int range, TMultiGraph* graph)
{
	int baseColor = kSpring;

	complete_data_header_type thisModelConfig = thisModel->getConfig();
	thisEmittanceAnalysis = new emittanceAnalysis( &thisModelConfig, thisModel );
	thisEmittanceAnalysis->calculateEmittanceContour(20);

	const int num_bunches = (thisModel->getDataAcquisition()).num_bunches;

	std::vector< double >* contourValues = thisEmittanceAnalysis->getContourValues();
	std::vector< double >* bunchEmittance; //Will point to values for bunch

	const int numberPoints = contourValues->size();

	std::vector<TGraph*> graphArray(num_bunches);

	for(int bunch = 0; bunch < num_bunches; ++bunch)
	{
		graphArray[bunch] = new TGraph();
		graphArray[bunch]->SetLineColor( baseColor + bunch % 20 - 10 );
		graphArray[bunch]->SetFillColor( baseColor + bunch % 20 - 10 );

		bunchEmittance = thisEmittanceAnalysis->getEmittanceCurves(bunch);

		for (int i=0; i < numberPoints; i++) graphArray[bunch]->SetPoint( i, (*contourValues)[i], (*bunchEmittance)[i] );

      	graph->Add(graphArray[bunch]);
	}

  	graph->SetTitle("Emittance");

  	delete thisEmittanceAnalysis;

}

void dataController::fillImageMomentPhaseHistogram(const int order, TH1F* hist)
{
	complete_data_header_type thisModelConfig = thisModel->getConfig();
	thisMomentAnalysis = new momentAnalysis();
	thisMomentAnalysis->setData( thisModel );

	thisMomentAnalysis->generateMomentMatrices(order);
	master->statusUpdate("moments generated");

	thisMomentAnalysis->calculateModeContent();
	master->statusUpdate("content calculated");

	const int num_bunches = thisModelConfig.dataAcquisition.num_bunches;

	//std::vector< double >* imageAmplitudes = thisEmittanceAnalysis->getImageMomentAmplitudes(order);
	hist->Clear();
	hist->SetBins( num_bunches, 1.0, (Double_t)num_bunches );
	hist->Scale(0.0);

	for(int b = 0; b < num_bunches; ++b)
	{
		double in    = thisMomentAnalysis->getInphase( order, b );
		double qu    = thisMomentAnalysis->getQuad( order, b );
		double phase = atan( qu / in );
		if( order == 0 )
		{
			phase = M_PI; // Phase is meaningless for m=0.
		}else if( order > 0 ){
			//For an m fold symmetric dist. only 2pi/m
			//phases are meaningful.
			phase += phase < 0.0 ? 2*M_PI : 0;
			while(phase > 2*M_PI / order) phase -= 2*M_PI/order;
			//phase = remainder( phase, 2*M_PI / order );
		}
		hist->SetBinContent( b+1, (float)phase );
	}

	char title[128];
	sprintf(title, "Order m=%d Phase", order);

	hist->SetTitle(title);
	int symmetry = order != 0 ? order : 1;
	hist->SetMaximum(  2*M_PI/symmetry );
	hist->SetMinimum( 0 );//-M_PI/symmetry
	hist->SetNdivisions(4,"Y");
	hist->SetFillColor(kAzure);

	delete thisMomentAnalysis;
	thisMomentAnalysis = NULL;

}

void dataController::fillImageMomentPhaseFFTHistogram(const int order, TH1F* hist)
{
	complete_data_header_type thisModelConfig = thisModel->getConfig();
	thisMomentAnalysis = new momentAnalysis();
	thisMomentAnalysis->setData( thisModel );

	thisMomentAnalysis->generateMomentMatrices(order);
	master->statusUpdate("moments generated");

	thisMomentAnalysis->calculateModeContent();
	master->statusUpdate("content calculated");

	const int num_bunches = thisModelConfig.dataAcquisition.num_bunches;

	std::vector<double> phases(num_bunches, 0.0);
	std::vector<double> fftVal(num_bunches, 0.0);

	for( int b = 0; b < num_bunches; ++b )
	{
		double in = thisMomentAnalysis->getInphase( order, b );
		double qu = thisMomentAnalysis->getQuad( order, b );
		phases[b] = atan( qu / in );
		if( order == 0 )
		{
			phases[b] = M_PI; // Phase is meaningless for m=0.
		}else if( order > 0 ){
			//For an m fold symmetric dist. only 2pi/m
			//phases are meaningful.
			phases[b] += phases[b] < 0.0 ? 2*M_PI : 0;
			//phases[b] = remainder( phases[b], 2*M_PI / order );
			phases[b] = fmod( phases[b], 2*M_PI / order );
		}
	}

	getArrayFFT( phases, thisModelConfig.dataAcquisition, fftVal );
	int fftSize = (int)ceil((double)num_bunches/2.0);

	hist->Clear();
	hist->SetBins( fftSize, 1.0, (Double_t)fftSize );
	hist->Scale(0.0);
	double fftNorm = 0.0;
	double re, im;
	std::vector<double> powerSpectrum(fftSize,0.0);
	for(int b = 0; b < fftSize; ++b)
	{
		re = fftVal[b];
		im = fftVal[b + fftSize ];
		powerSpectrum[b]  = re*re + im*im;
		fftNorm          += powerSpectrum[b];
	}
	fftNorm = fftNorm == 0.0 ? 1.0 : fftNorm;

	for(int b = 0; b < fftSize; ++b){ hist->SetBinContent( b+1, (float)powerSpectrum[b]/fftNorm ); }

	char title[128];
	sprintf(title, "Order m=%d Phase PowerSpectrum", order);

	hist->SetTitle(title);
	hist->SetMaximum(1.0);
	hist->SetMinimum(0.0);
	hist->SetFillColor(kOrange+7);

	delete thisMomentAnalysis;
	thisMomentAnalysis = NULL;

}

void dataController::fillDiscrepancyGraph( const int bunch, const int range, TMultiGraph* graph )
{
	//Using a float here even though disc is a double.
	// This will produced a 1D histogram of the final discrepancy values for each bunch.
	if(thisModel != NULL)
	{

		profile_data_header_type config = thisModel->getDataAcquisition();
		TGraph* thisGraph = new TGraph();
		for( int b = 0; b < config.num_bunches; ++b )
		{
			//Disc. is initialized to -1 since the value should
			//never be negative. So the first number less than one, or the end of
			//vector indicates the final value.
			int disc = 0;
			std::vector<double> *thisDisc = thisModel->getDiscVec(b);
			while( disc < thisDisc->size() && (*thisDisc)[disc] > 0 ) ++disc;
			double entry = (disc != 0) ? (*thisDisc)[disc-1] : 1.0; //if disc = 0 the while never evaluated. i.e. disc is empty
			thisGraph->SetPoint( b, (double)b, (double)entry );
		}

		graph->Add(thisGraph);
	}
}
void dataController::fillDiscrepancyHisto( const int bunch, const int range, TH1F* hist )
{
	//Using a float here even though disc is a double.
	// This will produced a 1D histogram of the final discrepancy values for each bunch.
	if(thisModel != NULL)
	{

		profile_data_header_type config = thisModel->getDataAcquisition();
		hist->SetBins( config.num_bunches, 0.0, (double)config.num_bunches );

		for( int b = 0; b < config.num_bunches; ++b )
		{
			//Disc. is initialized to -1 since the value should
			//never be negative. So the first number less than one, or the end of
			//vector indicates the final value.
			int disc = 0;
			std::vector<double> *thisDisc = thisModel->getDiscVec(b);
			while( disc < thisDisc->size() && (*thisDisc)[disc] > 0 ) ++disc;
			double entry = (disc != 0) ? (*thisDisc)[disc-1] : 1.0; //if disc = 0 the while never evaluated. i.e. disc is empty
			hist->SetBinContent( b, (float)entry );
		}
	}
}

void dataController::drawContour_NanoSec(const double d_nanosec, TVirtualPad* tvPad, const double xOffset )
{
	double BunchSpace = (1e+9)*1/(thisModel->getConfig()).machine.RF_f;
    double d_Phi = d_nanosec/BunchSpace * 2.0 * M_PI;

	drawContour_Phi( d_Phi, tvPad, xOffset );
}

void dataController::drawContour_MeV(const double d_MeV, TVirtualPad* tvPad, const double xOffset)
{
	double bucketHeightMEV  = (1e-6)*BucketHeight((thisModel->getConfig()).machine);
	double d_Phi = acos(1-2*(d_MeV*d_MeV)/(bucketHeightMEV*bucketHeightMEV));

	drawContour_Phi( d_Phi, tvPad, xOffset );
}

void dataController::drawContour_Phi( const double d_Phi, TVirtualPad* tvPad, const double xOffset )
{
	tvPad->cd();

	double BunchSpace       = (1e+9) * 1/(thisModel->getConfig()).machine.RF_f;
	double bucketHeightMEV  = (1e-6) * BucketHeight((thisModel->getConfig()).machine);

	double HalfRange = d_Phi   * BunchSpace/(2.0 * M_PI)*(1+0.01); // ten % padding gets the whole curve plotted
	double offset    = xOffset * BunchSpace/(2.0 * M_PI);

	TF1 *f1 = new TF1( "f1", "-1*[0]/1.41421*sqrt(-cos([2])-cos((x-[1]/2.0)*6.28318530718/[1]+[3])) ", -HalfRange + offset, HalfRange + offset );
    TF1 *f2 = new TF1( "f2", "   [0]/1.41421*sqrt(-cos([2])-cos((x-[1]/2.0)*6.28318530718/[1]+[3])) ", -HalfRange + offset, HalfRange + offset );

    f1->SetLineColor(46); f1->SetLineWidth(1);
    f1->SetParameters( bucketHeightMEV, BunchSpace, d_Phi, xOffset );
    f1->Draw("same");

    f2->SetLineColor(46); f2->SetLineWidth(1);
    f2->SetParameters( bucketHeightMEV, BunchSpace, d_Phi, xOffset );
    f2->Draw("same");
    tvPad->Update();
}

void dataController::drawRangeContours_Phi(const double d_Phi, TVirtualPad* pad, const int nBunches)
{
	if( nBunches == 1 )
	{
		drawContour_Phi(d_Phi, pad, 0.0);
	}else{
		for( int i = 0; i < nBunches; ++ i )
		{
			double offset = M_PI + i * 2 * M_PI;
			drawContour_Phi(d_Phi, pad, offset);
		}
	}
};

void dataController::drawRangeContours_MeV(const double d_MeV, TVirtualPad* pad, const int nBunches)
{
	if( nBunches == 1 )
	{
		drawContour_MeV(d_MeV, pad, 0.0);
	}else{
		for( int i = 0; i < nBunches; ++ i )
		{
			double offset = M_PI + i * 2 * M_PI;
			drawContour_MeV( d_MeV, pad, offset );
		}
	}
};

void dataController::drawRangeContours_NanoSec(const double d_nanosec, TVirtualPad* pad, const int nBunches)
{
	if( nBunches == 1 )
	{
		drawContour_NanoSec(d_nanosec, pad, 0.0);
	}else{
		for( int i = 0; i < nBunches; ++ i )
		{
			double offset = M_PI + i * 2 * M_PI;
			drawContour_NanoSec(d_nanosec, pad, offset);
		}
	}
};

void dataController::writeReconstruction(const char* filename)
{
	thisModel->writeImages(filename);
}

void dataController::padData(const int halfPad)
{
	thisModel->padData(halfPad);
}

complete_data_header_type dataController::getConfig()
{
	return thisModel->getConfig();
}

profile_data_header_type dataController::getDataAcquisition()
{
	return thisModel->getDataAcquisition();
}

Tomography_Config dataController::getTomography()
{
	return thisModel->getTomography();
}

Machine_Config dataController::getMachine()
{
	return thisModel->getMachine();
}

std::string dataController::dataInfo()
{
	std::string info;
	info = "info From dataController::dataInfo()";
	return info;
}

void dataController::setDataDataAquisition( const profile_data_header_type config )
{
	if( isNotEmpty() ) thisModel->setHeader( config );
}

void dataController::setDataTomography( const Tomography_Config config )
{
	if( isNotEmpty() ) thisModel->setTomography( config );
}

void dataController::setDataMachine( const Machine_Config config )
{
	if( isNotEmpty() ) thisModel->setMachine( config );
}

void dataController::fillDataFrom2DVector( const std::vector< std::vector<float> > input, const profile_data_header_type head )
{
	thisModel->readIn2DVector( input, head );
}
