#include "TomographyAnalysis.h"

double GetSynchPeriod(int order, Machine_Config machine){
    
    //Synch period calculated as 1/nu_s, only using values available from TomConf in GlobalConfig
    double gammaT  =  machine.gammaT;  
    double h       =  (double)machine.h;
    double eKin    =  machine.eKin;     
    double eV      =  machine.RF_V;    
    double mass    =  machine.mass;  
    double synPhi  =  machine.synPhi;

    double gamma = (eKin+mass)/mass;
    double beta  = sqrt(1-pow(gamma,-2.0));
    double eta   = pow(gammaT,-2.0)-pow(gamma,-2.0);

    // Estimate the "nominal" Synchrotron Period. tau has been replaced with 2PI*h/omegaRF
    // Sign of eta and cos(synPhi) should be +, make sure you're using the right synPhi
    double numerator   = -2*M_PI*pow(beta,2.0)*(eKin);
    double denominator =  eta*h*eV*cos( synPhi );

    double synchPeriod = sqrt(numerator/denominator);
    
    return synchPeriod/(double)order; //[turns]
}

double linearinterp(double x1, double x2, double y1, double y2, double y3){
    return (y1+x1*y1-x1*y2-y3)/(y1-y2);
}

double GuessConst(std::vector<double>& yValues){
    /// the constant offset is just the average value of the yData
    return std::accumulate(yValues.begin(), yValues.end(), 0.0)/(double)yValues.size();
}

double GuessPeriod(std::vector<double>& xValues,std::vector<double>& yValues,int order, Machine_Config machine){
    double meany = GuessConst(yValues);
    std::vector<int> zeroCrossPos1;
    double hpl;

    for(std::vector<int>::size_type i=0;i<yValues.size()-1;++i){ /// Find positions signal goes pos-neg or neg-pos
        //std::cout<<"yValues[i]-meany "<< yValues[i]-meany <<std::endl;
        if( ((yValues[i]-meany)<0&&(yValues[i+1]-meany)>0)||((yValues[i]-meany)>0&&(yValues[i+1]-meany)<0) ){
            zeroCrossPos1.push_back(i);
            //std::cout<<"zeroCrossPos1 i "<< i <<std::endl;
        }
    }

    if(zeroCrossPos1.size()>0){ /// i the data crosses y=0 then estimate cross positions

        std::vector<double> zeroCrossPos(zeroCrossPos1.size());
        for(std::vector<int>::size_type i=0;i<zeroCrossPos.size();++i){
            zeroCrossPos[i]=
            linearinterp(xValues[zeroCrossPos1[i]], xValues[zeroCrossPos1[i]+1], yValues[zeroCrossPos1[i]]-meany, yValues[zeroCrossPos1[i]+1]-meany, 0.0);
        }
        if(zeroCrossPos.size()>1){ /// i the data crosses y=0 then estimate cross positions
        
            hpl=zeroCrossPos[1]-zeroCrossPos[0];

            if(zeroCrossPos.size()>2){
                for(std::vector<int>::size_type i=1;i<zeroCrossPos.size()-1;++i){
                    hpl=(hpl+(zeroCrossPos[i+1]-zeroCrossPos[i]))/2;
                }
            }
        }else{
           hpl=GetSynchPeriod(order, machine); /// Half Synch period for current machine
        }
    }else{
           hpl=GetSynchPeriod(order, machine); /// Half Synch period for current machine
    }

 return 2*hpl;
}

void GuessAmpandPhase(std::vector<double>& xValues, std::vector<double>& yValues,std::vector<double>& ansVec,double perGuess){

    double minval = *min_element( yValues.begin(), yValues.end() );  /// The amp guess is half the max minus min values
    double maxval = *max_element( yValues.begin(), yValues.end() );
    int maxpos=0;                                           /// To guess the initial phase we calc the phase advance to the position of the Max value
    //std::cout << "yValues size = " << yValues.size() << std::endl;
    //This is segfauling often needs some error handling. 
    //It is trying to access values out of range in yValues. I think what's happening is 
    //that we're testing doubles! so they're might not be a match. Then you go out of range 
    //while(yValues[maxpos]!=maxval){ 
    //  ++maxpos;
    //}
    double current_max = minval;
    for(int i = 0; i < yValues.size(); ++i ){
        if(yValues[i] >= maxval ){
            maxpos = i;
        }
    }

    double phiGuess = 2*M_PI-(xValues[maxpos]*2*M_PI)/perGuess;  /// if we get a negative phase advance, add 2pi
    ansVec[0]=(maxval-minval)/2.0;  /// Fill answer vector
    ansVec[1]=phiGuess;
}

void FitDecayingCos(std::vector<double>& xValues,std::vector<double>& yValues,std::vector<double> & outVec,int order, Machine_Config machine){

    const double twoPi = 2*M_PI;
    /// Inital Guesses for FIT paarameters
    double cGuess = GuessConst( yValues );
    //std::cout<<"cGuess "<< cGuess <<std::endl;
    double perGuess = GuessPeriod( xValues, yValues, order, machine );
    //std::cout<<"perGuess "<< perGuess <<std::endl;
    std::vector<double> ampPhivec(2,0);
    GuessAmpandPhase(xValues,yValues,ampPhivec,perGuess);//perGuess);
    //std::cout<<"ampGuess "<< ampPhivec[0] <<std::endl;
    //std::cout<<"PhiGuess "<< ampPhivec[1] <<std::endl;
 
    outVec[5] = cGuess;
    outVec[6] = perGuess;
    outVec[7] = ampPhivec[0];
    outVec[8] = ampPhivec[1];

    /// create a graph object of the input {x,y} data
    TGraph *gr = new TGraph(xValues.size(),&xValues[0],&yValues[0]); /// Requires an Array type as input, therefore pass address of vector!

    /// This is the decaying cosine function we will try and model the data as
    TF1 *f = new TF1("f", "[4]+([0]*cos([1] + x*[2]))*TMath::Exp(x*[3])");

    f->SetParName(0, "amp");   /// More memorable names for fit parameters
    f->SetParName(1, "phi");
    f->SetParName(2, "omega");
    f->SetParName(3, "gamma");
    f->SetParName(4, "const");

    f->SetParameter(0, ampPhivec[0]);
    f->SetParLimits(0, -2.0*ampPhivec[0], 2.0*ampPhivec[0]); /// Setting some limits for the amplitude seems to help
                                                            /// We could set limits on other parameters if we want??

    f->SetParameter(1, ampPhivec[1]);
    f->SetParameter(2, twoPi/perGuess);
    f->SetParameter(3, -0.0004);
    f->SetParameter(4, cGuess);

    ///fit
    //std::cout<<"pp3 "<<std::endl;
    gr->Fit(f,"Q");

    //std::cout<<"pp4 "<<std::endl;
    //gr->Draw();


    /// assign the fit paramaters and errros to variables...
    TF1 *myfunc=gr->GetFunction("f");

    double amp0    = myfunc->GetParameter(0);
    double ampE0   = myfunc->GetParError(0);
    double phi0    = myfunc->GetParameter(1);
    double phiE0   = myfunc->GetParError(1);
    double omega0  = myfunc->GetParameter(2);
    double omegaE0 = myfunc->GetParError(2);
    double gamma0  = myfunc->GetParameter(3);
    double gammaE0 = myfunc->GetParError(3);
    double chi2    = myfunc->GetChisquare();

    //std::cout << "amp0  "<< amp0 << std::endl;
    //std::cout << "ampE0  "<< ampE0 << std::endl;

    /// add 2pi to Phi guess until it is positive...
    while(phi0<0){
        phi0 += twoPi;
    }
    /// sub 2pi to Phi guess until it is less than 2pi...
    while(phi0>twoPi){
        phi0 -= twoPi;
    }

    outVec[0] = amp0;
    outVec[1] = ampE0;
    outVec[2] = phi0-M_PI;
    outVec[3] = phiE0;
    outVec[4] = chi2;

   //   for(int k=0;k<outVec.size();k++)
   //     std::cout << "outVec  "<< outVec[k] << std::endl;
}

void GetMom1and2(std::vector<int>& vecIn, double *first, double *second){  
    /// Depending on how this is used it may be faster to calculate{
    /// all moments up to order n in one go
    double totVecIn;
    double mean     = 0.0;
    double variance = 0.0;

    totVecIn=(double)accumulate(vecIn.begin(),vecIn.end(), 0.0);

    for(std::vector<int>::size_type i=0; i<vecIn.size(); ++i){
        mean += (i+0.5)*vecIn[i]; /// this factor of a half is somewhat arbitrary, depends on definitions
    }
    mean=mean/totVecIn;

    for(std::vector<int>::size_type i=0; i<vecIn.size(); ++i){
        variance += (pow(i-mean,2))*vecIn[i];
    }
    *first  = mean;
    *second = sqrt(variance/totVecIn);
}

void GetMom1and2(std::vector<float>& vecIn, double *first, double *second){  
    /// Depending on how this is used it may be faster to calculate{
    /// all moments up to order n in one go
    double totVecIn;
    double mean     = 0.0;
    double variance = 0.0;

    totVecIn=(double)accumulate(vecIn.begin(),vecIn.end(), 0.0);

    for(std::vector<float>::size_type i=0; i<vecIn.size(); ++i){
        mean += (double)(i+0.5)*vecIn[i]; /// this factor of a half is somewhat arbitrary, depends on definitions
    }
    mean=mean/totVecIn;

    for(std::vector<float>::size_type i=0; i<vecIn.size(); ++i){
        variance += (pow(i-mean,2))*vecIn[i];
    }
    *first  = mean;
    *second = sqrt(variance/totVecIn);
}

void GetAveragedxData(std::vector<double>& vecOut, int turnspertrig,int startpos, int stoppos, int npToAverage){
    int numpos = (stoppos-startpos)/npToAverage;
    double turnspertrigd = (double)turnspertrig;
    std::vector<double> xValues(stoppos-startpos);

    for(int i=0; i<stoppos-startpos; ++i){
        xValues[i]=i*(double)turnspertrigd;
    }
    for(int i=0;i<numpos; ++i){
        vecOut[i]= (double)std::accumulate(xValues.begin()+startpos+i*npToAverage, xValues.begin()+startpos+i*npToAverage+npToAverage, 0.0)/(double)npToAverage;
    }
}

void GetAveragedyData(std::vector<double>& vecIn, std::vector<double>& vecOut, int startpos, int stoppos, int npToAverage){
    int numpos=(stoppos-startpos)/npToAverage;
    for(int k = 0; k<numpos; ++k){
        vecOut[k]= std::accumulate(vecIn.begin()+startpos+k*npToAverage, vecIn.begin()+startpos+k*npToAverage+npToAverage, 0.0)/npToAverage;
    }
}

void GetProfDipQuadMomData( std::vector<int>& profData, profile_data_header_type& profHead, std::vector<double>& dipOut,std::vector<double>& quadOut, Machine_Config machine){
  
    /// Estimate the "nominal" Synchrotron Period
    double synchPeriod = GetSynchPeriod(1, machine);
    //std::cout << "synchPeriod = " << synchPeriod << std::endl;

    /// Arbitrarily, take 1.75 synchPeriods of profiles (Unless that's too many)
    int numTurns = (int)floor(1.75*synchPeriod);///profHead.turns_per_trig);
    (numTurns > profHead.turns_per_trig*(profHead.num_frames-1) ? numTurns = profHead.turns_per_trig*(profHead.num_frames-1) : numTurns = numTurns );

    std::cout << " numTurns set = " << numTurns << std::endl;
    /// Vectors to fill with the dip and quad moments
    std::vector<double> dipMom(numTurns);
    std::vector<double> quadMom(numTurns);
    /// Temp variables
    std::vector<int> singleprof(numTurns);
    double a, b;

    int npToAverage = 5;  /// Assume dipMomVec data is noisey, so average of npToAverage pts will be taken
    /// xData For fitting, this is the same for each bunch
    std::vector<double> xDataForFit( (int)floor( (double)numTurns / (double)npToAverage ) );
    GetAveragedxData( xDataForFit, profHead.turns_per_trig, 0, numTurns, npToAverage );
    std::vector<double> yDataForFit(xDataForFit.size());     /// Set fill in loop;
    
    /// Vector/Matrixes for Dip and Quad answers
    std::vector<double> tempAns(9); 
    std::vector< std::vector<double> > dipAns(profHead.num_bunches ,tempAns);
    std::vector< std::vector<double> > quadAns(profHead.num_bunches,tempAns);

    for( int k = 0; k < profHead.num_bunches; ++k ) /// This will be a loop over the num frames per bunch
    {
        for( int j = 0; j < numTurns; ++j )
        {
            for( int i = 0; i < profHead.ptsperprofile; i++ )
            {
                singleprof[i] = profData[k*profHead.num_frames*profHead.ptsperprofile+j*profHead.ptsperprofile+i];
                //std::cout << k << " " << j << " " << singleprof[i] <<  std::endl;
            }
            GetMom1and2(singleprof,&a,&b);
            dipMom[j]  = a;
            quadMom[j] = b;
            //std::cout << j << " dipMom= " << dipMom[j] << " quadMom= " << quadMom[j] <<std::endl;
        }
        /// Get Dipole Moment averages and fit decaying Cos
        GetAveragedyData( dipMom, yDataForFit, 0, numTurns, npToAverage );
        FitDecayingCos( xDataForFit, yDataForFit, tempAns, 1, machine ); /// Fit
        
        for( int i = 0; i < tempAns.size(); ++i ){ dipAns[k][i] = tempAns[i]; }

        /// Get Quad Moment averages and fit decaying Cos
        GetAveragedyData( quadMom, yDataForFit, 0, numTurns, npToAverage );
        FitDecayingCos( xDataForFit, yDataForFit, tempAns, 2, machine ); /// Fit

        for( int i = 0; i < tempAns.size(); ++i){ quadAns[k][i] = tempAns[i]; }

        //for(i=0;i<1;++i)
        //std::cout << "dipAns " << k << " "<< dipAns[k][i] << "  quadAns[k] " << quadAns[k][i]  <<std::endl;
    }

    /// Fill answer vectors with desired paramaters
    for(int i=0;i<profHead.num_bunches;++i){
       for(int j=0;j<5;++j){ /// don't really like the hard-coded five...
       
           //std::cout << "i*5+j " << i*5+j <<std::endl;
           dipOut[i*5+j]  = dipAns[i][j];
           quadOut[i*5+j] = quadAns[i][j];
           //std::cout << "dipOut " << dipOut[i*5+j] << " quadOut " << quadOut[i*5+j]  <<std::endl;
       }

    }
}

void GetProfDipQuadMomData( std::vector<float>& profData, profile_data_header_type& profHead, std::vector<double>& dipOut,std::vector<double>& quadOut, Machine_Config machine){
  
    /// Estimate the "nominal" Synchrotron Period
    double synchPeriod = GetSynchPeriod(1, machine);
    //std::cout << "synchPeriod = " << synchPeriod << std::endl;

    /// Arbitrarily, take 1.75 synchPeriods of profiles (Unless that's too many)
    int numTurns = (int)floor(1.75*synchPeriod);///profHead.turns_per_trig);
    (numTurns > profHead.turns_per_trig * (profHead.num_frames - 1 )? numTurns = profHead.turns_per_trig * (profHead.num_frames -1): numTurns = numTurns );

    std::cout << " numTurns set = " << numTurns << std::endl;
    /// Vectors to fill with the dip and quad moments
    std::vector<double> dipMom(numTurns);
    std::vector<double> quadMom(numTurns);
    /// Temp variables
    std::vector<float> singleprof(profHead.ptsperprofile);
    double a, b;

    int npToAverage = 5;  /// Assume dipMomVec data is noisey, so average of npToAverage pts will be taken
    /// xData For fitting, this is the same for each bunch
    std::vector<double> xDataForFit( (int)floor( (double)numTurns / (double)npToAverage ) );
    GetAveragedxData( xDataForFit, profHead.turns_per_trig, 0, numTurns, npToAverage );
    std::vector<double> yDataForFit(xDataForFit.size());     /// Set fill in loop;
    
    /// Vector/Matrixes for Dip and Quad answers
    std::vector<double> tempAns(9); 
    std::vector< std::vector<double> > dipAns(profHead.num_bunches ,tempAns);
    std::vector< std::vector<double> > quadAns(profHead.num_bunches,tempAns);
    float dfloat;
    for( int k = 0; k < profHead.num_bunches; ++k ) /// This will be a loop over the num frames per bunch
    {
        for( int j = 0; j < numTurns; ++j )
        {
            for( int i = 0; i < profHead.ptsperprofile; i++ )
            {
                singleprof[i] = profData[ k * profHead.num_frames * profHead.ptsperprofile + j * profHead.ptsperprofile + i ];
                //std::cout << k << " " << j << " " << singleprof[i] <<  std::endl;
            }
            GetMom1and2(singleprof,&a,&b);
            dipMom[j]  = a;
            quadMom[j] = b;
            //std::cout << j << " dipMom= " << dipMom[j] << " quadMom= " << quadMom[j] <<std::endl;
        }
        /// Get Dipole Moment averages and fit decaying Cos
        GetAveragedyData( dipMom, yDataForFit, 0, numTurns, npToAverage );
        FitDecayingCos(   xDataForFit, yDataForFit, tempAns, 1, machine ); /// Fit
        
        for( int i = 0; i < tempAns.size(); ++i ){ dipAns[k][i] = tempAns[i]; }

        /// Get Quad Moment averages and fit decaying Cos
        GetAveragedyData( quadMom, yDataForFit, 0, numTurns, npToAverage );
        FitDecayingCos( xDataForFit, yDataForFit, tempAns, 2, machine ); /// Fit

        for( int i = 0; i < tempAns.size(); ++i){ quadAns[k][i] = tempAns[i]; }

        //for(i=0;i<1;++i)
        //std::cout << "dipAns " << k << " "<< dipAns[k][i] << "  quadAns[k] " << quadAns[k][i]  <<std::endl;
    }

    /// Fill answer vectors with desired paramaters
    for(int i=0;i<profHead.num_bunches;++i){
       for(int j=0;j<5;++j){ /// don't really like the hard-coded five...
       
           //std::cout << "i*5+j " << i*5+j <<std::endl;
           dipOut[i*5+j]  = dipAns[i][j];
           quadOut[i*5+j] = quadAns[i][j];
           //std::cout << "dipOut " << dipOut[i*5+j] << " quadOut " << quadOut[i*5+j]  <<std::endl;
       }

    }
}

double Get1stMom(std::vector<double>& vecIn)  /// Depending on how this is used it may be faster to calculate
{
    std::vector<int>::size_type i;
    double totVecIn = (double)accumulate(vecIn.begin(),vecIn.end(), 0.0);
    double mean     = 0.0;
    double variance = 0.0;
    
    for(i=0; i < vecIn.size(); ++i){
        mean += (i+0.5)*vecIn[i]; /// this factor of a half is somewhat arbitrary, depends on definitions
    }
    mean=mean/totVecIn;

    return mean;
}

double SimpleTrackForDipAmp(double xMom,double yMom ,profile_data_header_type& profFileHeader, Machine_Config machine){
    int i;
    int k;

    /// Machine Parameters
    const double harm   = machine.h;
    const double fRF    = machine.RF_f;
    const double vRF    = machine.RF_V;
    const double Ekin   = machine.eKin;
    const double mp     = machine.mass;
    const double gammaT = machine.gammaT;
    
    const double beamp=(harm*(pow(gammaT,-2.0) - (pow(mp,2.0)/pow(Ekin + mp,2.0))))/((Ekin*fRF*(Ekin + 2*mp))/(Ekin + mp));

   /// these are the bunch widths in units of eV and seconds

    const double Tbin= 1/(fRF*profFileHeader.ptsperprofile);
    const double Ebin=(2*sqrt(Ekin*(Ekin + 2*mp))*sqrt(2/M_PI)*sqrt(((mp + Ekin)*vRF)/harm))/((Ekin + mp)*profFileHeader.ptsperprofile*sqrt(fabs(-(pow(mp,2.0)/pow(Ekin + mp,2.0)) + pow(gammaT,-2.0))));

    double Tin;
    double Ein;
    double Enew;
    double Tnew;
    double maxT = 0.0;


    //for(k=0;k<xMom.size();++k)
   // {
        Tin = xMom*Tbin;
        Ein = yMom*Ebin;
        maxT=0.0;
        for(int i=1; i<1000; ++i)
        {

            Ein=Ein+vRF*sin(2.0*M_PI*Tin*fRF);
            Tin=Tin+beamp*Ein;

            if(maxT<fabs(Tin)) maxT=fabs(Tin);
            
        }
     return maxT/Tbin;
    //}
}

void GetReconDipMom(std::vector<float>& reconData, profile_data_header_type& profFileHeader, double *amp,  double *phase, Machine_Config machine){
    
    double hbl = profFileHeader.ptsperprofile/2.0;
    /// Temporary vectors during calculation

    std::vector<double> xProj(profFileHeader.ptsperprofile);
    std::vector<double> yProj(profFileHeader.ptsperprofile);
    double xMom;
    double yMom;

    /// Project on to horizontal and vertical
        for(int j=0; j<xProj.size(); ++j){
            yProj[j]=0;
            for(int l=0; l<profFileHeader.ptsperprofile; ++l){
                yProj[j]+=(double)reconData[j*profFileHeader.ptsperprofile+l];
            }
        }

        for(int l=0; l<xProj.size(); l++){
            xProj[l]=0;
            for(int j=0; j<profFileHeader.ptsperprofile; j++){
                xProj[l]+=(double)reconData[l+j*profFileHeader.ptsperprofile];
            }
        }

        /// Get (x,y) moments about "nominal" bucket centre
        xMom = Get1stMom(xProj)-hbl;
        yMom = Get1stMom(yProj)-hbl;

    /// Phases are simple...
    *phase = atan(yMom/xMom)-M_PI/2.0f;
    /// Simple tracking to find max amplitude in time
    *amp   = SimpleTrackForDipAmp( xMom, yMom, profFileHeader, machine);
}

void GetProfDipFFT(std::vector<double> &profDipMomData,  profile_data_header_type& profHead, std::vector<double> &vecOut){
    /// We need to get the phase part of the input vector
    std::vector<double> profDipPhases(profHead.num_bunches);   /// Fill with fitted phases
    for(int i=0;i<profDipPhases.size();++i)
    {
        profDipPhases[i]=profDipMomData[i*5+2];
    }
    int n = profDipPhases.size()-1;
    int m =(int)ceil((float)n/2.0);
    TH1D *hist = new TH1D("hist", "hist", n, -4, 4); /// range is actuall -pi to pi for phases, does this matter???
    for(int i=0; i<profHead.num_bunches; i++)                     ///Fill the histogram with vector values
    {
        hist->SetBinContent(i,profDipPhases[i]);
    }
    /// This does the FFT and fills rePart imPart, copied from root turtorial...
    TH1 *hm = 0;
    TVirtualFFT::SetTransform(0);
    hm = hist->FFT(hm, "MAG");
    TVirtualFFT *fft = TVirtualFFT::GetCurrentTransform();
    /// Fill vecOut
    fft->GetPointsComplex((Double_t*)&vecOut[0],(Double_t*)&vecOut[m]);

    delete hist;
}

void GetProfQuadFFT(std::vector<double> &profQuadMomData,  profile_data_header_type& profHead,std::vector<double> &vecOut)
{
    /// We need to get the phase part of the input vector
    std::vector<double> profQuadPhases(profHead.num_bunches);   /// Fill with fitted phases
    for(int i=0;i<profQuadPhases.size();++i)
    {
        profQuadPhases[i]=profQuadMomData[i*2+1];
    }
    int n = profQuadPhases.size()-1;
    int m = (int)ceil((float)n/2.0);
    TH1D *hist = new TH1D("hist", "hist", n, -4, 4); /// range is actuall -pi to pi for phases, does this matter???
    for(int i=0; i<profHead.num_bunches; i++)                     ///Fill the histogram with vector values
    {
        hist->SetBinContent(i,profQuadPhases[i]);
    }
    /// This does the FFT and fills rePart imPart, copied from root turtorial...
    TH1 *hm =0;
    TVirtualFFT::SetTransform(0);
    hm = hist->FFT(hm, "MAG");
    TVirtualFFT *fft = TVirtualFFT::GetCurrentTransform();
    /// Fill vecOut
    fft->GetPointsComplex((Double_t*)&vecOut[0],(Double_t*)&vecOut[m]);

    delete hist;
    delete hm;
}

void getArrayFFT(const std::vector<double> &input, const profile_data_header_type& profHead, std::vector<double> &output)
{
    if(input.size() == profHead.num_bunches)
    {
        int n = profHead.num_bunches - 1;
        int m = (int)ceil((float)n/2.0);
        TH1D *hist = new TH1D("hist", "hist", n, -4, 4); /// range is actuall -pi to pi for phases, does this matter???
        for(int i=0; i < profHead.num_bunches; i++)      ///Fill the histogram with vector values
        {
            hist->SetBinContent( i, input[i] );
        }
        /// This does the FFT and fills rePart imPart, copied from root turtorial...
        TH1 *hm = 0;
        TVirtualFFT::SetTransform(0);
        hm = hist->FFT(hm, "MAG");
        TVirtualFFT *fft = TVirtualFFT::GetCurrentTransform();

        /// Fill vecOut
        output.resize(m);
        fft->GetPointsComplex((Double_t*)&output[0], (Double_t*)&output[m]);

        delete hist;
        delete hm;
    }
}

void GetProfFFTMags( std::vector<double> &profDipMomData, std::vector<double> &profQuadMomData, profile_data_header_type& profHead, std::vector<double> &dipDataOut, std::vector<double> &quadDataOut){

    std::vector<double> profDipFFTTemp(  profHead.num_bunches );
    std::vector<double> profQuadFFTTemp( profHead.num_bunches );


    GetProfDipFFT(  profDipMomData , profHead, profDipFFTTemp );
    GetProfQuadFFT( profQuadMomData, profHead, profQuadFFTTemp);

    unsigned int n = dipDataOut.size();

    for(int i=0;i<n;++i){
        dipDataOut[i]  = sqrt( pow( profDipFFTTemp[i]  , 2 ) + pow( profDipFFTTemp[n+i] , 2) );
        quadDataOut[i] = sqrt( pow( profQuadFFTTemp[i] , 2 ) + pow( profQuadFFTTemp[n+i], 2) );
    }
}

//Original versions of the functions overloaded to take GlobalConfigs are below. 

double GetSynchPeriod(int order)
{
    //This needs to be replaced with a TomConf structure or something. 
    const double cVac     = 299792458;

    const double circumMI = 531.24;
    const double gammaT   = 21.6;
    const double harm     = 588.0;
    const double Ekin     = 8.0*pow(10.0,9.0);
    const double fRF      = 52811400.0;
    const double vRF      = 150000.0;
    const double mp       = 9.3827231*pow(10.0,8.0);
    
    //const double gammaT = 26.1;
    
    /// Estimate the "nominal" Synchrotron Period
    double synchPeriod=(sqrt((cVac*fRF*vRF)/(circumMI*sqrt(Ekin*(Ekin + mp))))*sqrt(fabs(pow(gammaT,-2)-pow(mp,2)/pow(Ekin + mp,2))))/(2.*M_PI);
    std::cout << "Duncans period = " << synchPeriod/(double)order << std::endl;
    return synchPeriod/(double)order;
}

double GuessPeriod(std::vector<double>& xValues,std::vector<double>& yValues,int order)
{
    double meany=GuessConst(yValues);
    std::vector<int> zeroCrossPos1;
    double hpl;

    for(std::vector<int>::size_type i=0;i<yValues.size()-1;++i){ /// Find positions signal goes pos-neg or neg-pos
        //std::cout<<"yValues[i]-meany "<< yValues[i]-meany <<std::endl;
        if( ((yValues[i]-meany)<0&&(yValues[i+1]-meany)>0)||((yValues[i]-meany)>0&&(yValues[i+1]-meany)<0) ){
            zeroCrossPos1.push_back(i);
            //std::cout<<"zeroCrossPos1 i "<< i <<std::endl;
        }
    }

    if(zeroCrossPos1.size()>0){ /// i the data crosses y=0 then estimate cross positions

        std::vector<double> zeroCrossPos(zeroCrossPos1.size());
        for(std::vector<int>::size_type i=0;i<zeroCrossPos.size();++i){
            zeroCrossPos[i]=
            linearinterp(xValues[zeroCrossPos1[i]], xValues[zeroCrossPos1[i]+1], yValues[zeroCrossPos1[i]]-meany, yValues[zeroCrossPos1[i]+1]-meany, 0.0);
        }
        if(zeroCrossPos.size()>1){ /// i the data crosses y=0 then estimate cross positions
        
            hpl=zeroCrossPos[1]-zeroCrossPos[0];

            if(zeroCrossPos.size()>2){
                for(std::vector<int>::size_type i=1;i<zeroCrossPos.size()-1;++i){
                    hpl=(hpl+(zeroCrossPos[i+1]-zeroCrossPos[i]))/2;
                }
            }
        }else{
           hpl=GetSynchPeriod(order); /// Half Synch period for 150kV 53MhzRF
        }
    }else{
           hpl=GetSynchPeriod(order); /// Half Synch period for 150kV 53MhzRF
    }

 return 2*hpl;
}

void FitDecayingCos(std::vector<double>& xValues,std::vector<double>& yValues,std::vector<double> & outVec,int order)
{
    const double twoPi= 2*M_PI;
    /// Inital Guesses for FIT paarameters
    double cGuess=GuessConst(yValues);
    //std::cout<<"cGuess "<< cGuess <<std::endl;
    double perGuess=GuessPeriod(xValues,yValues,order);
    //std::cout<<"perGuess "<< perGuess <<std::endl;
    std::vector<double> ampPhivec(2);
    GuessAmpandPhase(xValues,yValues,ampPhivec,perGuess);//perGuess);
    //std::cout<<"ampGuess "<< ampPhivec[0] <<std::endl;
    //std::cout<<"PhiGuess "<< ampPhivec[1] <<std::endl;

    outVec[5]=cGuess;
    outVec[6]=perGuess;
    outVec[7]=ampPhivec[0];
    outVec[8]=ampPhivec[1];

    /// create a graph object of the input {x,y} data
    TGraph *gr = new TGraph(xValues.size(),&xValues[0],&yValues[0]); /// Requires an Array type as input, therefore pass address of vector!

    /// This is the decaying cosine function we will try and model the data as
    TF1 *f = new TF1("f", "[4]+([0]*cos([1] + x*[2]))*TMath::Exp(x*[3])");

    f->SetParName(0, "amp");   /// More memorable names for fit parameters
    f->SetParName(1, "phi");
    f->SetParName(2, "omega");
    f->SetParName(3, "gamma");
    f->SetParName(4, "const");

    f->SetParameter(0, ampPhivec[0]);
    f->SetParLimits(0, -2.0*ampPhivec[0],2.0*ampPhivec[0]); /// Setting some limits for the amplitude seems to help
                                                            /// We could set limits on other parameters if we want??

    f->SetParameter(1,  ampPhivec[1]);
    f->SetParameter(2, twoPi/perGuess);
    f->SetParameter(3,-0.0004);
    f->SetParameter(4, cGuess);

    ///fit
    //std::cout<<"pp3 "<<std::endl;
    gr->Fit(f,"Q");

    //std::cout<<"pp4 "<<std::endl;
    //gr->Draw();


    /// assign the fit paramaters and errros to variables...
    TF1 *myfunc=gr->GetFunction("f");

    double amp0    = myfunc->GetParameter(0);
    double ampE0   = myfunc->GetParError(0);
    double phi0    = myfunc->GetParameter(1);
    double phiE0   = myfunc->GetParError(1);
    double omega0  = myfunc->GetParameter(2);
    double omegaE0 = myfunc->GetParError(2);
    double gamma0  = myfunc->GetParameter(3);
    double gammaE0 = myfunc->GetParError(3);
    double chi2    = myfunc->GetChisquare();

    //std::cout << "amp0  "<< amp0 << std::endl;
    //std::cout << "ampE0  "<< ampE0 << std::endl;

    /// add 2pi to Phi guess until it is positive...
    while(phi0<0){
        phi0+=twoPi;
    }
    /// sub 2pi to Phi guess until it is less than 2pi...
    while(phi0>twoPi){
        phi0-=twoPi;
    }

    outVec[0]=amp0;
    outVec[1]=ampE0;
    outVec[2]=phi0-3.14156;
    outVec[3]=phiE0;
    outVec[4]=chi2;

   //   for(int k=0;k<outVec.size();k++)
   //     std::cout << "outVec  "<< outVec[k] << std::endl;
}

void GetProfDipQuadMomData( std::vector<int>& profData,profile_data_header_type& profHead, std::vector<double>& dipOut,std::vector<double>& quadOut)
{
  
    /// Estimate the "nominal" Synchrotron Period
    double synchPeriod=GetSynchPeriod(1);

    /// Arbitrarily, take 1.5 synchPeriods of profiles (Unless that's too many)
    int numTurns= (int)floor(1.5*synchPeriod);
    (numTurns > profHead.turns_per_trig * profHead.num_frames ? numTurns = profHead.turns_per_trig * profHead.num_frames : numTurns=numTurns );

    /// Vectors to fill with the dip and quad moments
    std::vector<double> dipMom(numTurns);
    std::vector<double> quadMom(numTurns);
    /// Temp variables
    std::vector<int> singleprof(numTurns);
    double a, b;

    int npToAverage=5;  /// Assume dipMomVec data is noisey, so average of npToAverage pts will be taken

    /// xData For fitting, this is the same for each bunch
    std::vector<double> xDataForFit((int)floor((double)numTurns/(double)npToAverage));
    GetAveragedxData(xDataForFit, profHead.turns_per_trig,0,numTurns,npToAverage);
    std::vector<double> yDataForFit(xDataForFit.size());     /// Set fill in loop;

    /// Vector/Matrixes for Dip and Quad answers
    std::vector<double> tempAns(9);
    std::vector<std::vector<double> > dipAns(profHead.num_bunches,tempAns);
    std::vector<std::vector<double> > quadAns(profHead.num_bunches,tempAns);

    for(int k=0;k<profHead.num_bunches;++k) /// This will be a loop over the num frames per bunch
    {
        for(int j=0;j<numTurns;++j)
        {
            for(int i=0;i<profHead.ptsperprofile;i++)
            {
                singleprof[i]=profData[k*profHead.num_frames*profHead.ptsperprofile+j*profHead.ptsperprofile+i];
                //std::cout << k << " " << j << " " << singleprof[i] <<  std::endl;
            }
            GetMom1and2(singleprof,&a,&b);
            dipMom[j]=a;
            quadMom[j]=b;
            //std::cout << j << " dipMom= " << dipMom[j] << " quadMom= " << quadMom[j] <<std::endl;
        }
        /// Get Dipole Moment averages and fit decaying Cos
        GetAveragedyData(dipMom,yDataForFit,0,numTurns,npToAverage);
        FitDecayingCos(xDataForFit,yDataForFit,tempAns,1); /// Fit
        for(int i=0;i<tempAns.size();++i)
        dipAns[k][i]=tempAns[i];
        /// Get Quad Moment averages and fit decaying Cos
        GetAveragedyData(quadMom,yDataForFit,0,numTurns,npToAverage);
        FitDecayingCos(xDataForFit,yDataForFit,tempAns,2); /// Fit
        for(int i=0;i<tempAns.size();++i)
        quadAns[k][i]=tempAns[i];
        //for(i=0;i<1;++i)
        //std::cout << "dipAns " << k << " "<< dipAns[k][i] << "  quadAns[k] " << quadAns[k][i]  <<std::endl;
    }

    /// Fill answer vectors with desired paramaters
    for(int i=0;i<profHead.num_bunches;++i){
       for(int j=0;j<5;++j){ /// don't really like the hard-coded five...
       
           //std::cout << "i*5+j " << i*5+j <<std::endl;
           dipOut[i*5+j]=dipAns[i][j];
           quadOut[i*5+j]=quadAns[i][j];
           //std::cout << "dipOut " << dipOut[i*5+j] << " quadOut " << quadOut[i*5+j]  <<std::endl;
       }

    }
}

double SimpleTrackForDipAmp(double xMom,double yMom ,profile_data_header_type& profFileHeader)
{
    int i;
    int k;

    /// Machine Parameters
    const double harm     = 588.0;
    const double fRF      = 52811400.0;
    const double vRF      = 150000.0;
    const double circumMI = 531.24;
    const double cVac     = 299792458;
    const double Ekin     = 8.0*pow(10.0,9.0);
    const double mp       = 9.3827231*pow(10.0,8.0);
    const double gammaT   = 21.6;

    const double beamp=(harm*(pow(gammaT,-2.0) - (pow(mp,2.0)/pow(Ekin + mp,2.0))))/((Ekin*fRF*(Ekin + 2*mp))/(Ekin + mp));

   /// these are the bunch widths in units of eV and seconds

    const double Tbin= 1/(fRF*profFileHeader.ptsperprofile);
    const double Ebin=(2*sqrt(Ekin*(Ekin + 2*mp))*sqrt(2/M_PI)*sqrt(((mp + Ekin)*vRF)/harm))/((Ekin + mp)*profFileHeader.ptsperprofile*sqrt(fabs(-(pow(mp,2.0)/pow(Ekin + mp,2.0)) + pow(gammaT,-2.0))));

    double Tin;
    double Ein;
    double Enew;
    double Tnew;
    double maxT=0.0;


    //for(k=0;k<xMom.size();++k)
   // {
        Tin = xMom*Tbin;
        Ein = yMom*Ebin;
        maxT=0.0;
        for(i=1;i<1000;++i)
        {
            Ein=Ein+vRF*sin(2.0*M_PI*Tin*fRF);
            Tin=Tin+beamp*Ein;
            if(maxT<fabs(Tin))
            {
                maxT=fabs(Tin);
            }
        }
     return maxT/Tbin;
    //}
}

void GetReconDipMom(std::vector<float>& reconData,profile_data_header_type& profFileHeader,double *amp,  double *phase)
{
    int i,j,k,l;
    double hbl=profFileHeader.ptsperprofile/2.0;
    /// Temporary vectors during calculation

    std::vector<double> xProj(profFileHeader.ptsperprofile);
    std::vector<double> yProj(profFileHeader.ptsperprofile);
    double xMom;
    double yMom;

    /// Project on to horizontal and vertical
        for(j=0; j<xProj.size(); ++j)
        {
            yProj[j]=0;
            for(l=0; l<profFileHeader.ptsperprofile; ++l)
            {
                yProj[j]+=(double)reconData[j*profFileHeader.ptsperprofile+l];
            }

        }
        for(l=0; l<xProj.size(); l++)
        {
            xProj[l]=0;
            for(j=0; j<profFileHeader.ptsperprofile; j++)
            {
                xProj[l]+=(double)reconData[l+j*profFileHeader.ptsperprofile];
            }
        }
        /// Get (x,y) moments about "nominal" bucket centre
        xMom=Get1stMom(xProj)-hbl;
        yMom=Get1stMom(yProj)-hbl;

    /// Phases are simple...
    *phase=atan(yMom/xMom)-M_PI/2.0f;
    /// Simple tracking to find max amplitude in time
    *amp=SimpleTrackForDipAmp(xMom,yMom,profFileHeader);
}
