#include "momentAnalysis.h"

momentAnalysis::momentAnalysis()
{
	currentOrder = -1; // negative means no matrices.
	normalMatrices.clear();
	skewedMatrices.clear();

	inphase.clear();
	quad.clear();

	assocData = NULL;
}

momentAnalysis::~momentAnalysis()
{
	;
}

void momentAnalysis::generateMomentMatrices( int upToOrder, int coordinates )
{
	double dPhiMax = M_PI / 2.0;
	int upTo = upToOrder == 0 ? 1 : upToOrder;
	if( coordinates == kCartesian && hasData() )
	{
		Machine_Config machine = assocData->getMachine();

		int nx = ( assocData->getConfig() ).tomography.pixelsPhi; 
		int ny = ( assocData->getConfig() ).tomography.pixelsdE ; 

		nx = nx <= 0 ? ( assocData->getConfig() ).dataAcquisition.ptsperprofile : nx; 
		ny = nx <= 0 ? ( assocData->getConfig() ).dataAcquisition.ptsperprofile : ny; 

		double xLo = 0.00;
		double xHi = (double)nx;
		double yLo = 0.00;
		double yHi = (double)ny;

		normalMatrices.clear(); //inphase components
		skewedMatrices.clear(); //quadrature components

		double bunchSpacingNSec = 1.0E9 / machine.RF_f;       //[ns]
		double maxRad_Time = dPhiMax / M_PI * bunchSpacingNSec / 2.0; //[ns]
		double maxRad_Bins = maxRad_Time * 1.0 / ( assocData->getConfig() ).dataAcquisition.hscale; //[ns/ns]

		double yScale = calculateYScale( dPhiMax, nx, ny );

		double centerX = 0.5;
		double centerY = 0.5;

		for( int i = 0; i <= upTo; ++i )
		{	
			char histoTitle[128];
			sprintf(histoTitle, "%f %f", yScale, maxRad_Time);
			normalMatrices.push_back( new TH2D("","", nx, xLo, xHi, ny, yLo, yHi ) ); //inphase components
			skewedMatrices.push_back( new TH2D("","", nx, xLo, xHi, ny, yLo, yHi ) ); //quadrature components

			generateOrderM( i, normalMatrices[i], kNormal, maxRad_Bins, yScale, centerX, centerY ); //Get rid of nx/2.0 here.
			generateOrderM( i, skewedMatrices[i], kSkewed, maxRad_Bins, yScale, centerX, centerY ); //Get rid of nx/2.0 here.
		}

		currentOrder = upTo;

	}else if( coordinates == kPolar && hasData() ){
		;//No polar coordinates for now. Maybe later if you want.
	}
}

void momentAnalysis::calculateModeContent()
{
	if( currentOrder >= 0 && hasData() )
	{
		inphase.clear();
		quad.clear();
		int numBunches    = ( assocData->getDataAcquisition() ).num_bunches;

		int pixelsPhi = ( assocData->getConfig() ).tomography.pixelsPhi; 
		int pixelsdE  = ( assocData->getConfig() ).tomography.pixelsdE ; 

		pixelsPhi = pixelsPhi <= 0 ? ( assocData->getConfig() ).dataAcquisition.ptsperprofile : pixelsPhi; 
		pixelsdE  = pixelsdE  <= 0 ? ( assocData->getConfig() ).dataAcquisition.ptsperprofile : pixelsdE ; 

		for( int b = 0; b < numBunches; ++b )
		{
			float *bunch = assocData->getImage( b );
			for( int m = 0; m <= currentOrder; ++m )
			{
				double normalValue = 0.0;
				double skewedValue = 0.0;

				for( int p = 0; p < pixelsPhi * pixelsdE; ++p )
				{
					double imageVal = (double) *(bunch + p);
					normalValue += imageVal * normalMatrices[m]->GetBinContent( p + 1 );
					skewedValue += imageVal * skewedMatrices[m]->GetBinContent( p + 1 );
				}

				inphase.push_back( normalValue );
				quad.push_back( skewedValue );
			}
		}
	}
}

void momentAnalysis::extractPhases( std::vector<double> phaseStorage )
{
	phaseStorage.clear();
	double phase;
	for( int i = 0; i < inphase.size(); ++i )
	{
		phase = atan( quad[i]/inphase[i] );
		phaseStorage.push_back( phase );
	}
}
//Center is given as a faction of the image width. The default 0.5 has the moment span the image
//this is not a good idea, so it should be set. 
// maxRad is effectively the phi semi-axis
// yScale is the relative scaling factor to be given to the y axis. i.e. phiSemiAxis/dESemiAxis
void momentAnalysis::generateOrderM( int m, TH2D* hist, int normORskew, double maxRad, double yScale, double centerX, double centerY )
{
	double weight, xVal, yVal, rad, rToMinusM;

	int nBinsX = hist->GetNbinsX();
	int nBinsY = hist->GetNbinsY();

	for( int x = 0; x < nBinsX; ++x )
	{
		xVal = x + 0.5 - nBinsX * centerX;
		for( int y = 0; y < nBinsY; ++y )
		{	
			yVal = y + 0.5 - nBinsY * centerY;
			yVal *= yScale;
			rad  = sqrt( pow( xVal, 2 ) + pow( yVal, 2 ) );
			weight    = 0.0;
			rToMinusM = 1.0/( pow( rad, m ) );
			if( rad <= 0.0 ) rToMinusM = 0.0;
			if( rad <= maxRad )
			{	
				// Because: ( x + iy )^m = SumOverK( (m choose k)*x^(m-k)*(iy)^k )
				// and
				// i^0=1, i^1=i, i^2=-1, i^3=-i, i^4=1 ... step by four with j, get both Re or Im at once
				// We have:
				// 		k%4 = 0,2 are Re w/ sign (+,-) respectively 
				// 		k%4 = 1,3 are Im w/ sign (+,-) respectively 
				for( int j = 0; j <= m; j += 4 ) 
				{
					int k = ( normORskew == kNormal ) ? j : j + 1; 
					if( k <= m ) weight += mChooseK( m, k ) * pow(xVal, m-k ) * pow( yVal, k );
					k += 2;
					if( k <= m ) weight -= mChooseK( m, k ) * pow(xVal, m-k ) * pow( yVal, k );
				}
			}	
			hist->SetCellContent( x+1 , y+1 , rToMinusM * weight );
		}
	}

	// f(r,T) = Rn(r)*Sum( Am*Cos(mT) + Bm*Sin(mT) ) over all m
	//Integral( f(r,T) * Cos(nT), {0,2*Pi} ) = Am*Pi  if n==m!=0, Am*2Pi if n==m==0, 0 otherwise
	//So scale m==0 component by 1/2. Since Sin(0T)==0 no problem for out of phase component.
	//if( m == 0 ) hist->Scale( M_PI / hist->Integral() );
			 
}

double momentAnalysis::calculateYScale( double dPhiMax, int nBinsX, int nBinsY )
{
	//This will calculate the scaling needed to get a proper elipse
	//for multiplication with image. The image will be multiplied bin by bin,
	//not using histos of images
	//It is fine to leave everything in terms of bin number.
	//It is assumed the bunch is centered in y. If it's not, too bad.

	Machine_Config machine = assocData->getMachine();

	double bucketHeightBins = (double)nBinsY/2.0;
	double heightInBins     = sqrt( 1 - cos( dPhiMax ) ) * bucketHeightBins / sqrt(2.0);  //top of contour
	double bunchSpacingNSec = 1.0E9 / machine.RF_f;
	double widthInTime      = dPhiMax/M_PI * bunchSpacingNSec;
	double widthInBins      = widthInTime * 1.0 / ( assocData->getConfig() ).dataAcquisition.hscale;

	double yScale = heightInBins / ( widthInBins / 2.0 );

	return yScale;
}

void momentAnalysis::printRawMomentEllipsesToFile()
{
	if( currentOrder >= 0 )
	{
		TCanvas* c1 = new TCanvas();
		char title[128];

		for( int i = 0; i < normalMatrices.size(); ++i )
		{
			c1->Clear();
			c1->Size(5.0,5.0);
			c1->cd();
			sprintf( title, "/usr/local/cbs_files/cns_write/pa4219/Trash/normal_%d.pdf", i );
			normalMatrices[i]->Draw("COLZ");
			c1->Print(title);

			c1->Clear();
			c1->Size(5.0,5.0);
			c1->cd();
			sprintf( title, "/usr/local/cbs_files/cns_write/pa4219/Trash/skewed_%d.pdf", i );
			skewedMatrices[i]->Draw("COLZ");
			c1->Print(title);
		}
		delete c1;
	}

}

double momentAnalysis::getInphase(int order, int bunch)
{
	double val = 0.0;
	int getPos = (currentOrder+1) * bunch + order;
	if( order <= currentOrder && inphase.size() > getPos ) val = inphase[ getPos ];
	return val;
}

double momentAnalysis::getQuad(int order, int bunch)
{
	double val = 0.0;
	int getPos = (currentOrder+1) * bunch + order;
	if( order <= currentOrder && quad.size() > getPos ) val = quad[ getPos ];
	return val;
}

int momentAnalysis::factorial( int x )
{
	int ans = 1;
	for(int i = x; i > 1; --i) ans *= i;
	return ans;
}

TH2D* momentAnalysis::getMthOrderMatrix(int order, int orientation)
{
	TH2D* hPtr = NULL;
	if( orientation == kNormal && order <= currentOrder )
	{
		hPtr = normalMatrices[order];
	}else if( orientation == kNormal && order <= currentOrder ){
		hPtr = skewedMatrices[order]; 
	}
	return hPtr;
}


double momentAnalysis::mChooseK( int m, int k )
{
	double ans = -1;
	int num;
	int denom;
	if( m >= k )
	{
		num   = factorial(m);
		denom = factorial(m-k) * factorial(k);
		ans = (double)num/(double)denom;
	}

	return ans;
}

