//Map class
#include "MapClass.h"

//N.B. At map creation time it is unknown which maps will be used.
//because of this the bins are numbered locally from 1,ptsperframe
//during reconstruction bins are numbered globally from 1,ptsperframe*numberOfFrames
//because of this reading from the fiMap, or map files into the 
//vector BinV must take into account the global bin position.
//An additional ptsperframe*Frame must be added.

MapClass::MapClass( const complete_data_header_type init, const std::string& path, const int type, const double center )
{
	//Map path should be set at creation time using some external string. 

	MapPath   = path; 
  mapLength = 0 ; //Right now vectors are all empty.
  mapCenter = center; //map center should be given as a bin? or fraction of image? Probably the latter.
  //Negative values mean default to one half the image width.
  //center defaults to -1.0

  copyProfile_data_header_type( &init.dataAcquisition, &header.dataAcquisition );
  copyTomography( &init.tomography, &header.tomography );  
  copyMachine( &init.machine, &header.machine );

  //Initializing a map object necessarily creates appropriate maps.
  //Map types:
  //  Standard  - start at t=0 make maps for every turn to t=Tmax
  //  Sparse    - start at t=0 make maps for every nth turn to t=Tmax (n=turnsPerTrigger)
  //  Extended  - start at t=0 make maps for every turn to t=+/-Tmax
  //  SingleUse - same as Standard but stored in memory. deleted with maps object. All other written to file.
  //  N.B. + combintions. i.e. SparseSingleUse stores sparse maps in memory. etc.

  if( mapsExist() )//mapsExist() sets timestamp if true.
  {  
    #ifdef _STANDALONE_VERBOSE
      std::cout << "Maps exist." << std::endl;
    #endif

  	readInMaps(); //Uses timestamp. 

  }else{
    #ifdef _STANDALONE_VERBOSE
      std::cout << "Maps don't exist." << std::endl;
    #endif

    if( type == kIGOR )
    {
      igorsMaps();
    }else{
      createMaps( type );//Sets timestamp
    }

    if( type == kStandard || type == kSparse || type == kSparseExtended || type == kExtended ) readInMaps(); //Uses timestamp
    //No need to read in if single use.
  }

}

MapClass::~MapClass()
{
	;
}

void MapClass::printHeader(){ print_complete_data_header_type( header ); }
void MapClass::printDataAcquisition(){ print_profile_data_header_type( header.dataAcquisition ); }
void MapClass::printTomography(){ print_Tomography_Config( header.tomography ); }
void MapClass::printMachine(){ print_Machine_Config( header.machine ); }

void MapClass::setTomography(const Tomography_Config newTomography){ copyTomography( &newTomography, &header.tomography ); }
void MapClass::setMachine(const Machine_Config newMachine){ copyMachine( &newMachine, &header.machine ); }

void MapClass::readInMaps()
{
  mapLength = 0; //Always reset to zero when reading in.

  std::ifstream MapFile;
  std::stringstream MapFileName;
  
  int mapEntry[3];
  size_t entrySize  = sizeof(mapEntry);
  int ptsperprofile = header.dataAcquisition.ptsperprofile;

  #ifdef _STANDALONE_VERBOSE
    std::cout << "timestamp = " << timestamp << std::endl;
  #endif

  PixV.clear();
  BinV.clear();
  ValV.clear();

  for( int frame = 0; frame < header.tomography.Frames; ++frame )
  {

    int offSetFrame = frame - header.tomography.ReconstructionFrame;   
    MapFileName << MapPath << "/" << timestamp << "_0_To_" << offSetFrame * header.dataAcquisition.turns_per_trig << ".map"; 

    #ifdef _DEBUG_MAPS
      std::cout << "Looking for map " << MapFileName.str() << std::endl;
    #endif

    MapFile.open( MapFileName.str().c_str(), std::ios::binary | std::ios::in );
  
    if( !MapFile.is_open() )
    { 
      PixV.clear();
      BinV.clear();
      ValV.clear();
      #ifdef _STANDALONE_VERBOSE
        std::cout << "Maps Not Found." <<std::endl; 
      #endif
      exit(12);
    }
    
    // while( !feof( MapFile ) )
    while( MapFile.read( (char *)&mapEntry, entrySize ) )
    {
		  PixV.push_back( mapEntry[0] );
		  BinV.push_back( mapEntry[1] + ptsperprofile*frame ); //See N.B. at top of file.
		  ValV.push_back( mapEntry[2] );
      ++mapLength;
    }

    MapFile.close();
    MapFileName.str("");
    MapFileName.clear();
  }

  fillNSubJ();

  #ifdef _STANDALONE_VERBOSE
    std::cout << "Maps Read In." <<std::endl; 
  #endif
}

bool MapClass::mapsExist()
{
  bool mapsExist = false;

  char Index[256];
  sprintf( Index, "%s/%s", MapPath.c_str(), "ExistingMapsIndex.ind");

  complete_data_header_type oldHeader;
  time_t oldTimestamp;

  FILE* IndexFile = fopen( Index, "r" );
  	
  if(IndexFile == NULL)
  {
		timestamp = -1;
	}else{

    while( !feof(IndexFile) && !mapsExist )
    {
      fscanf( IndexFile, "%ld", &oldTimestamp );
      //Data Acq
      fscanf( IndexFile, "%f", &oldHeader.dataAcquisition.hscale         );
      fscanf( IndexFile, "%d", &oldHeader.dataAcquisition.turns_per_trig );
      //Tomography
      fscanf( IndexFile, "%d", &oldHeader.tomography.MaxIterations );
      fscanf( IndexFile, "%d", &oldHeader.tomography.Frames        );
      fscanf( IndexFile, "%d", &oldHeader.tomography.SqrtPPP       );
      fscanf( IndexFile, "%d", &oldHeader.tomography.pixelsPhi     );
      fscanf( IndexFile, "%d", &oldHeader.tomography.pixelsdE      );
      //Machine
      fscanf( IndexFile, "%d" , &oldHeader.machine.h      );
      fscanf( IndexFile, "%lf", &oldHeader.machine.eKin   );
      fscanf( IndexFile, "%lf", &oldHeader.machine.mass   );
      fscanf( IndexFile, "%lf", &oldHeader.machine.RF_V   );
      fscanf( IndexFile, "%lf", &oldHeader.machine.RF_f   );
      fscanf( IndexFile, "%lf", &oldHeader.machine.gammaT );
      fscanf( IndexFile, "%lf", &oldHeader.machine.synPhi );

      if( mapsExist = mapsSuffice( oldHeader ) ) timestamp = oldTimestamp;
    }

  	fclose( IndexFile );
  }
	
  return mapsExist;
};

bool MapClass::mapsSuffice( const complete_data_header_type oldHeader )
{
	//Parameters that don't effect maps are left out.
	//This should be made smarter when you get around
	//To making the back projection smart. 
	//Maps should be made up to a synch period by default.
	//Even if given a lower value. If higher overwrite. 
	//Since those maps will be sufficient.
	//CurConfig should be the configuration you're running.
	//This matters for Turns. If you run 200 turns and a map
	//has been made for 300, no need to remake maps.

	bool sufficientQ     = false;
  int roundedWidth_Cur = (int)floor( 100 * oldHeader.dataAcquisition.hscale + 0.5 );
  int roundedWidth_Old = (int)floor( 100 * header.dataAcquisition.hscale    + 0.5 );
  if(
    ( roundedWidth_Cur                      == roundedWidth_Old                         )&&
    ( header.dataAcquisition.turns_per_trig == oldHeader.dataAcquisition.turns_per_trig )&&
		
		( header.tomography.Frames  <= oldHeader.tomography.Frames  )&& 
		( header.tomography.SqrtPPP == oldHeader.tomography.SqrtPPP )&&
    ( header.tomography.pixelsPhi == oldHeader.tomography.pixelsPhi )&&
    ( header.tomography.pixelsdE  == oldHeader.tomography.pixelsdE  )&&

		( header.machine.h      == oldHeader.machine.h      )&&
		( header.machine.eKin   == oldHeader.machine.eKin   )&&
		( header.machine.mass   == oldHeader.machine.mass   )&&
		( header.machine.RF_V   == oldHeader.machine.RF_V   )&&
		( header.machine.RF_f   == oldHeader.machine.RF_f   )&&
		( header.machine.gammaT == oldHeader.machine.gammaT )&&
		( header.machine.synPhi == oldHeader.machine.synPhi )
    ) sufficientQ = true;

	return sufficientQ;
};

void MapClass::logMap()
{
  
	char IndexFileName[512];
	sprintf( IndexFileName, "%s/%s", MapPath.c_str(), "ExistingMapsIndex.ind");

  FILE* IndexFile = fopen( IndexFileName ,"a" );

  //write timestamp
  fprintf( IndexFile, "%ld\n", timestamp );
  //Data Acq
  fprintf( IndexFile, "%f\n", header.dataAcquisition.hscale         );
  fprintf( IndexFile, "%d\n", header.dataAcquisition.turns_per_trig );
  //Tomography
  fprintf( IndexFile, "%d\n", header.tomography.MaxIterations );
  fprintf( IndexFile, "%d\n", header.tomography.Frames        );
  fprintf( IndexFile, "%d\n", header.tomography.SqrtPPP       );
  fprintf( IndexFile, "%d\n", header.tomography.pixelsPhi     );
  fprintf( IndexFile, "%d\n", header.tomography.pixelsdE      );
  //Machine
  fprintf( IndexFile, "%d\n" , header.machine.h      );
  fprintf( IndexFile, "%lf\n", header.machine.eKin   );
  fprintf( IndexFile, "%lf\n", header.machine.mass   );
  fprintf( IndexFile, "%lf\n", header.machine.RF_V   );
  fprintf( IndexFile, "%lf\n", header.machine.RF_f   );
  fprintf( IndexFile, "%lf\n", header.machine.gammaT );
  fprintf( IndexFile, "%lf\n", header.machine.synPhi );

  //close file
  fclose( IndexFile );
};

void MapClass::createMaps( const int type )
{
  bool isSparse            ;
  bool shouldTrackBackwards;
  bool shouldWriteToFile   ;

  switch(type)
  {
    case kStandard:
      isSparse             = false;
      shouldTrackBackwards = false;
      shouldWriteToFile    = true;
    break;
    case kExtended:
      isSparse             = false;
      shouldTrackBackwards = true;
      shouldWriteToFile    = true;
    break;
    case kSparse:
      isSparse             = true;
      shouldTrackBackwards = false;
      shouldWriteToFile    = true;
    break;
    case kSparseExtended:
      isSparse             = true;
      shouldTrackBackwards = true;
      shouldWriteToFile    = true;
    break;
    case kSparseSingleUse:
      isSparse             = true;
      shouldTrackBackwards = false;
      shouldWriteToFile    = false;
    break;
  }

  timestamp = time(NULL);
  
  int start;
  int end;
  int TimeDummy = 0;
  int TimeTrack = 0;
  int TimeFill  = 0;

  start = clock();

  float tPi = 2*M_PI;

  //Machine
  double eKin   = header.machine.eKin;
  double mass   = header.machine.mass;
  double eV     = header.machine.RF_V;   //eV
  double fRF    = header.machine.RF_f;   //Hz
  double gammaT = header.machine.gammaT;
  double synPhi = header.machine.synPhi;

  int h = header.machine.h;

  #ifdef _DEBUG_MAPS
    std::cout << "Header in maps: "<< std::endl;
    printHeader();
    std::cout << std::endl;
    std::cout << "Machine in maps: "<< std::endl;
    printMachine();
    std::cout << std::endl;
    std::cout << "Tomography in maps: "<< std::endl;
    printTomography();
    std::cout << std::endl;
  #endif

  //Tomography
  //NEED SOME WAY TO CHANGE THE BUNCH CENTER.
  double centerFraction = ( mapCenter < 0.0 || mapCenter > 1.0 ) ? 0.5 : mapCenter; 
  double BunchCenter = header.dataAcquisition.ptsperprofile/2.0; //This should maybe be settable? 
  double BinWidth    = header.dataAcquisition.hscale; //ns
  
  int TurnsPerTrig = header.dataAcquisition.turns_per_trig;
  int Turns        = header.tomography.Frames * TurnsPerTrig;

  int turnsForMapping = shouldTrackBackwards ? 2*Turns      : Turns ; 
  int turnIncrement   =             isSparse ? TurnsPerTrig : 1     ;

  //Setting up pixel, bin sizes etc.
  int pSize        = header.dataAcquisition.ptsperprofile;  // profiles size 
  int iSize        = header.tomography.pixelsPhi;//header.dataAcquisition.ptsperprofile;  //  image x size
  int jSize        = header.tomography.pixelsdE ;//iSize;                                 //  image y size

  iSize = iSize <= 0 ? pSize : iSize; // size defaults to size of profiles if not explicitly set.
  jSize = jSize <= 0 ? pSize : jSize; // size defaults to size of profiles if not explicitly set.

  float  imagePhiCenter = (float)iSize / (float)2.0;
  float  imagedECenter  = (float)jSize / (float)2.0;
  double pixWidth    = BinWidth * (double)pSize / (double)iSize;

  int Pixels       = iSize*jSize;
  int SqrtPPCell   = header.tomography.SqrtPPP;

  //Derived
  float phiParam   = slip( header.machine )*tPi*h / ( pow( beta( header.machine ),2) * es( header.machine ) );
  float sPhiS      = sin( synPhi );
  float bHeight    = BucketHeight( header.machine );
  float TimeWidth  = (1/fRF)*pow(10.0,9.0);  //in ns since BinWidth given in ns
  float BunchBins  = TimeWidth/BinWidth; //Width of bunch in bins.
  double BunchPix  = TimeWidth/pixWidth;//Width of bunch in pixels
  float phiBin     = tPi/(float)pSize; //tPi/(float)iSize; 
  float phiCell    = tPi/(float)iSize;
  float dECell     = 2*bHeight/float(jSize-1);  

  std::vector<float> coords(SqrtPPCell,0);

  // Coordinates of grid in a single cell. 
  // Shift the coordinates by the cell number over all cells and track 
  for( int i = 0; i < SqrtPPCell ; i++ ) coords[ i ] = ( 2*float(i)+ 1)/( 2*float(SqrtPPCell) );
  
  int IsOut;
  int Pix;
  std::vector<int> DummyPhis(Turns,0);
  std::vector<int> DummydEs(Turns,0);

  float Phi,dE;
  //These are used for binning in phi
  float PhiLoLimit = -BunchCenter        * phiBin;
  float PhiUpLimit = (pSize-BunchCenter) * phiBin;

  //Tests will test if the particle is in the separatrix
  //This test should probably be done a little better.
  //This certainly could lead to errors. Things outside, but near
  //enough to the separatrix to not leave the image range over the course
  //of the tracking will contribute in some way. 
  //In fact, this is probably the error seen in signals with artificial
  //cable dispersion.
  float PhiLoTest  = (imagePhiCenter - BunchPix/2.0) * phiCell - M_PI; //Test used to ask if particle is in tracking area.
  float PhiUpTest  = (imagePhiCenter + BunchPix/2.0) * phiCell - M_PI; //Test used to ask if particle is in tracking area.
  
  std::vector< Level2Map > fiMap (turnsForMapping);

  if( -M_PI > PhiLoTest ) PhiLoTest = -M_PI;
  if(  M_PI < PhiUpTest ) PhiUpTest =  M_PI;

  #ifdef _STANDALONE_VERBOSE
    std::cout << "Creating maps." << std::endl;
  #endif

  for( int bini = 0; bini < iSize; ++bini )
  {
    for( int binj = 0; binj < jSize; ++binj )
    {
      Pix = bini + binj * iSize; //No need to recalculate in sub bin loop.
      for( int subi = 0; subi < SqrtPPCell; ++subi )
      {
       for( int subj = 0; subj < SqrtPPCell; ++subj )
       {
          //Positions given as phi,dE relative to BunchCenter
          //i.e. bunch center is at 0,0
          IsOut = 0;
          Phi = (coords[subi] + bini - imagePhiCenter) * phiCell;
          dE  = (coords[subj] + binj - imagedECenter) * dECell ;

          //________Track Turn__________________________
          //Forward Tracking
          for( int ThisTurn = 0; ( ThisTurn < Turns )&&( IsOut == 0 ); ++ThisTurn )
          {
            //Make this test better. 
            if( (PhiLoTest > Phi) || (PhiUpTest < Phi) )IsOut = 1; //Still in phi?
            if( (-bHeight > dE)   || (bHeight < dE)    )IsOut = 1; //Still in dE?

            //This needs to be updated... how?
            DummyPhis[ThisTurn] = (int)floor( (Phi-PhiLoLimit)/phiBin ); // We use phiBin because we're binning for the projections  

            //dummy = next_dphi_dE( dummy, eV, sPhiS, phiParam );
            dE  += eV * (sin(Phi) - sPhiS);
            Phi += dE * phiParam;
          }//Array of bins for 1 particle each turn.

          //_______One Tracked_________________________

          //TimeDummy = clock();
          if( !IsOut )
          {
            for( int TrnInd = 0; TrnInd < Turns; TrnInd += turnIncrement )
            {
              fiMap[ TrnInd ][ Pix ][ DummyPhis[TrnInd] ] += 1;  
            }
          }

          //BackwardTracking for extended map types
          if( shouldTrackBackwards && !IsOut )
          {
            //________Track Turn Backward___________________
            //IsOut = 0; //This shouldn't get reset maybe? Also a little error from not removing if we go out backward.
            //Shouldn't happen though if stationary bucket. 
            Phi = (coords[subi] + bini - imagePhiCenter) * phiCell; // We use cell values here becuase we're populating pixels.
            dE  = (coords[subj] + binj - imagedECenter ) * dECell ; // We use cell values here becuase we're populating pixels.
      
            for( int ThisTurn = -1; ( ThisTurn > -Turns )&&( IsOut == 0 ); --ThisTurn )
            {
              if( (PhiLoTest > Phi) || (PhiUpTest < Phi) ) IsOut = 1; //Still in phi?
              if( (-bHeight > dE)   || (bHeight < dE)    ) IsOut = 1; //Still in dE?

              // First track, then bin. This makes sure that the first entry is turn -1.
              // Turn 0 was handled in the forward tracking.  
              Phi -= dE*phiParam;
              dE  -= eV*(sin(Phi)-sPhiS); 
              //Turn -1 goes in part 0! Be Careful!
              DummyPhis[ -ThisTurn-1 ] = (int)floor( (Phi-PhiLoLimit)/phiBin ); // We use bin value here becuase we're binning for projections.
              //DummydEs[  -ThisTurn-1 ] = (int)floor( (dE+bHeight)/dEBin      ); 

            }//Array of bins for 1 particle each turn.

            //_______One Tracked_________________________

            //TimeDummy = clock();
            //Here the map uses turns > turns to store the negative turns information
            //So (1,2,3...Turns,-1,-2,-3...-Turns) This could alternatively be done
            // with a separate map for the -t turns. 
            if( !IsOut )
            {
              for( int TrnInd = Turns; TrnInd < 2*Turns; TrnInd += turnIncrement )
              {
                fiMap[ TrnInd ][ Pix ][ DummyPhis[ TrnInd % Turns ] ] += 1;  
              }
            }
          }//End BackwardTracking

          //TimeFill += clock() - TimeDummy;
        }//subj
      }//subi    
    }//binj
    #ifdef _LOUD_MAPS
      std::cout << "Row "<< bini <<" of "<< iSize << " tracked." << std::endl;
    #endif
  }//bini

  end = clock();
  #ifdef _STANDALONE_VERBOSE
    std::cout << "Maps Created." << std::endl;
  #endif 
  //std::cout << "Tracking complete: "<< (float)(end-start)/CLOCKS_PER_SEC << " sec." << std::endl;
  //std::cout << "\tTracking: "<< (float)(TimeTrack)/CLOCKS_PER_SEC << " sec." << std::endl;
  //std::cout << "\tFilling:  "<< (float)(TimeFill)/CLOCKS_PER_SEC << " sec." << std::endl;

  if( shouldWriteToFile )
  { 
    writeMapFiles( fiMap, type );
    logMap(); //_Update Log 
    //function for reading should fillnSubJals() else it won't be filled!
  }else{
    mapToLocalVectors( fiMap, type );
    fillNSubJ();
  }

  #ifdef _LOUD_MAPS
    std::cout << "mapLength = " << mapLength << std::endl;

    std::cout << "PixV.size() = " << PixV.size() << std::endl;
    std::cout << "BinV.size() = " << BinV.size() << std::endl;
    std::cout << "ValV.size() = " << ValV.size() << std::endl;
  #endif
};//End TrackForMaps

void MapClass::mapToLocalVectors( const std::vector< Level2Map > &fiMap, const int type )
{
  //Default we need Turns worth of maps and we need them every turn.
  int Turns      = header.tomography.Frames * header.dataAcquisition.turns_per_trig;
  int mapTurns   = Turns;
  int increment  = 1;
  int ptsperprofile = header.dataAcquisition.ptsperprofile;

  //If request is for: 
  //  sparse   -need to write maps every turnsPerTrig turns.  
  switch(type)
  {
    case kSparseSingleUse:
      increment = header.dataAcquisition.turns_per_trig;
    break;
    default:
    break;
  }

  std::map<int, int>::const_iterator kIt;
  std::map<int, Level1Map>::const_iterator iIt;

  PixV.clear();
  BinV.clear();
  ValV.clear();

  mapLength = 0;
  int frame = 0; //vector index is continous, TurnIndex may jump by turnsPerTrigger 
  for( int TrnInd = 0; TrnInd < Turns; TrnInd += increment )
  { 
    for( iIt = (fiMap[TrnInd]).begin(); iIt != (fiMap[TrnInd]).end(); ++iIt )
    {
      for( kIt = (*iIt).second.begin(); kIt != (*iIt).second.end(); ++kIt )
      {
        PixV.push_back((*iIt).first );
        BinV.push_back((*kIt).first + ptsperprofile*frame );
        ValV.push_back((*kIt).second);
        ++mapLength;
      }
    }

    #ifdef _LOUD_MAPS
      std::cout << "Frame " << frame << " of maps populated." << std::endl;
    #endif

    ++frame;
  }


}

void MapClass::writeMapFiles( const std::vector< Level2Map > &fiMap, const int type )
{
  //Default we need Turns worth of maps and we need them every turn.
  int Turns      = header.tomography.Frames * header.dataAcquisition.turns_per_trig;
  int mapTurns   = Turns;
  int increment  = 1;

  //If request is for: 
  //  sparse   -need to write maps every turnsPerTrig. 
  //  extended -need to write twice as many maps. 
  switch(type)
  {
    case kSparse:
      increment = header.dataAcquisition.turns_per_trig;
    break;
    case kSparseExtended:
      increment = header.dataAcquisition.turns_per_trig;
      mapTurns *= 2;
    break;
    case kExtended:
      mapTurns *=2;
    break;
    default:
    break;
  }

  std::ofstream     dummyMapFileStream;
  std::stringstream dummyFileName;

  int realTurn;    //Used in fileName. Convert loop index to actual turn number
  int mapEntry[3];
  size_t entrySize = sizeof( mapEntry );

  std::map<int, int>::const_iterator kIt;
  std::map<int, Level1Map>::const_iterator iIt;
   
  for( int TrnInd = 0; TrnInd < mapTurns; TrnInd += increment )
  {
    //TrnInd greater than 'Turns' reserved for backward maps
    realTurn = TrnInd < Turns ? TrnInd : -(TrnInd-Turns)-increment;

    #ifdef _LOUD_MAPS
      std::cout <<  " Writing map " << realTurn << " for TrnInd = " << TrnInd << std::endl;
    #endif

    //Deal with this a little more elegantly. Yuck.
    #ifndef _WINDOWS_MACHINE
      dummyFileName << MapPath << "/" << timestamp << "_0_To_" << realTurn << ".map";
    #endif

    #ifdef _WINDOWS_MACHINE
      dummyFileName << MapPath << "\\" << timestamp << "_0_To_" << realTurn << ".map";
    #endif

    dummyMapFileStream.open( dummyFileName.str().c_str(), std::ios::out | std::ios::binary );

    for( iIt = (fiMap[TrnInd]).begin(); iIt != (fiMap[TrnInd]).end(); ++iIt )
    {
      for( kIt = (*iIt).second.begin(); kIt != (*iIt).second.end(); ++kIt )
      {
        mapEntry[0] = (*iIt).first;   //Pix
        mapEntry[1] = (*kIt).first;   //Bin
        mapEntry[2] = (*kIt).second;  //Val
        dummyMapFileStream.write( (char*)&mapEntry, entrySize );
      }
    }
    dummyMapFileStream.close();
    dummyFileName.str("");
    dummyFileName.clear();
  }
}

void MapClass::dumpMapsToStdOut()
{
  // for(int frame = 0; frame < PixV.size(); ++frame)
  // {
    // std::cout <<"Maps for frame " << frame << ":" << std::endl; 
    // for(int j = 0; j < PixV[frame].size(); ++j)
    // {
    //   std::cout << " PixV = " << PixV[frame][j] << " BinV = " << BinV[frame][j] << " ValV = " << ValV[frame][j] << std::endl;
    // }
  //}
}

void MapClass::fillNSubJ()
{
  //We don't use any of these except BinTot. Can probably get rid of the other two.
  nSubJ.resize( header.tomography.Frames * header.dataAcquisition.ptsperprofile , 0 );
  BinTot.resize( header.tomography.Frames * header.dataAcquisition.ptsperprofile, 0 );
  PixTot.resize( header.dataAcquisition.ptsperprofile * header.dataAcquisition.ptsperprofile, 0 );

  int bin;
  for( int entry = 0; entry < BinV.size(); ++entry )
  { 
      bin = BinV[ entry ];
      ++nSubJ[ bin ];
      BinTot[ bin ] += ValV[ entry ];
      //PixTot[ PixV[ entry ] ] += ValV[ entry ];
  }
}

void MapClass::igorsMaps()
{
  bool isSparse             = false;
  bool shouldTrackBackwards = false;
  bool shouldWriteToFile    = false;

  timestamp = time(NULL);
  
  int start;
  int end;
  int TimeDummy = 0;
  int TimeTrack = 0;
  int TimeFill  = 0;

  start = clock();

  float tPi = 2*M_PI;

  //Machine
  double eKin   = header.machine.eKin;
  double mass   = header.machine.mass;
  double eV     = header.machine.RF_V;   //eV
  double fRF    = header.machine.RF_f;   //Hz
  double gammaT = header.machine.gammaT;
  double synPhi = header.machine.synPhi;

  int h = header.machine.h;

  #ifdef _DEBUG_MAPS
    std::cout << "Header in maps: "<< std::endl;
    printHeader();
    std::cout << std::endl;
    std::cout << "Machine in maps: "<< std::endl;
    printMachine();
    std::cout << std::endl;
    std::cout << "Tomography in maps: "<< std::endl;
    printTomography();
    std::cout << std::endl;
  #endif

  //Tomography
  //NEED SOME WAY TO CHANGE THE BUNCH CENTER.
  double centerFraction = ( mapCenter < 0.0 || mapCenter > 1.0 ) ? 0.5 : mapCenter; 
  double BunchCenter = header.dataAcquisition.ptsperprofile/2.0; //This should maybe be settable? 
  double BinWidth    = header.dataAcquisition.hscale; //ns
  
  int TurnsPerTrig = header.dataAcquisition.turns_per_trig;
  int Turns        = header.tomography.Frames * TurnsPerTrig;
  int iSize        = header.dataAcquisition.ptsperprofile;
  int jSize        = iSize;
  int Pixels       = iSize*iSize;
  int SqrtPPCell   = header.tomography.SqrtPPP;

  int turnsForMapping = shouldTrackBackwards ? 2*Turns : Turns; 
  int turnIncrement   = isSparse ? TurnsPerTrig : 1 ;

  //Derived
  float phiParam  = slip( header.machine )*tPi*h / ( pow( beta( header.machine ),2) * es( header.machine ) );
  float sPhiS     = sin( synPhi );
  float bHeight   = BucketHeight( header.machine );
  float dEBin     = 2*bHeight/float(jSize-1);
  float TimeWidth = (1/fRF)*pow(10.0,9.0);  //in ns since BinWidth given in ns
  float BunchBins = TimeWidth/BinWidth; //Width of bunch in bins.
  float phiBin    = tPi/(float)iSize;   

  std::vector<float> coords(SqrtPPCell,0);

  //coordinates of grid in a single cell
  for( int i = 0; i < SqrtPPCell ; i++ ) coords[ i ] = (2*float(i)+1)/(2*float(SqrtPPCell));
  // shift the coordinates by the cell number over all cells and track 
  
  int IsOut;
  int Pix;
  std::vector<int> DummyPhis(Turns,0);
  std::vector<int> DummydEs(Turns,0);

  float Phi,dE;
  //Limits are the limit of the image.
  float PhiLoLimit = -BunchCenter*phiBin;
  float PhiUpLimit = (iSize-BunchCenter)*phiBin;

  //Tests will test if the particle is in the separatrix
  //This test should probably be done a little better.
  //This certainly could lead to errors. Things outside, but near
  //enough to the separatrix to not leave the image range over the course
  //of the tracking will contribute in some way. 
  //In fact, this is probably the error seen in signals with artificial
  //cable dispersion.
  float PhiLoTest  = (BunchCenter-BunchBins/2.0)*phiBin - M_PI; //PhiLoLimit;
  float PhiUpTest  = (BunchCenter+BunchBins/2.0)*phiBin - M_PI; //PhiUpLimit;
    
  std::vector< Level2Map > fiMap (turnsForMapping);

  if( -M_PI > PhiLoTest ) PhiLoTest = -M_PI;
  if(  M_PI < PhiUpTest ) PhiUpTest =  M_PI;

  #ifdef _STANDALONE_VERBOSE
    std::cout << "Creating maps." << std::endl;
  #endif

  for( int bini = 0; bini < iSize; ++bini )
  {
    for( int binj = 0; binj < jSize; ++binj )
    {
      Pix = bini + binj*iSize; //No need to recalculate in sub bin loop.
      for( int subi = 0; subi < SqrtPPCell; ++subi )
      {
       for( int subj = 0; subj < SqrtPPCell; ++subj )
       {
          //Positions given as phi,dE relative to BunchCenter
          //i.e. bunch center is at 0,0
          IsOut = 0;
          Phi = (coords[subi] + bini - BunchCenter) * phiBin;
          dE  = (coords[subj] + binj - BunchCenter) * dEBin ;

          //________Track Turn__________________________
          //Forward Tracking
          for( int ThisTurn = 0; ( ThisTurn < Turns )&&( IsOut == 0 ); ++ThisTurn )
          {
            //Make this test better. 
            if( (PhiLoTest > Phi) || (PhiUpTest < Phi) )IsOut = 1; //Still in phi?
            if( (-bHeight > dE)   || (bHeight < dE)    )IsOut = 1; //Still in dE?

            //This needs to be updated... how?
            DummyPhis[ThisTurn] = (int)floor( (Phi-PhiLoLimit)/phiBin ); 
            DummydEs[ThisTurn]  = (int)floor( (dE+bHeight)/dEBin      ); 

            //dummy = next_dphi_dE( dummy, eV, sPhiS, phiParam );
            dE  += eV*(sin(Phi) - sPhiS);
            Phi += dE*phiParam;
          }//Array of bins for 1 particle each turn.

          //_______One Tracked_________________________

          //TimeDummy = clock();
          if( !IsOut )
          {
            //This was as mistake. You really just want the bin number.
            //You need to take the binning from 1,m into accound when reading in the maps
            //since m is dependent on frame.
            // int frame = 0; 
            for( int TrnInd = 0; TrnInd < Turns; TrnInd += turnIncrement )
            {
              // fiMap[ TrnInd ][ Pix ][ DummyPhis[TrnInd] + frame * iSize ] += 1;  
              // ++frame;
              fiMap[ TrnInd ][ Pix ][ DummyPhis[TrnInd] ] += 1;  
            }
          }

          //BackwardTracking for extended map types
          if( shouldTrackBackwards && !IsOut )
          {
            //________Track Turn Backward___________________
            IsOut = 0; //This shouldn't get reset maybe? Also a little error from not removing if we go out backward.
            //Shouldn't happen if though if stationary bucket. 
            Phi = (coords[subi] + bini - BunchCenter)*phiBin;
            dE  = (coords[subj] + binj - BunchCenter)*dEBin;
      
            for( int ThisTurn = -1; ( ThisTurn > -Turns )&&( IsOut == 0 ); --ThisTurn )
            {
              if( (PhiLoTest > Phi) || (PhiUpTest < Phi) ) IsOut = 1; //Still in phi?
              if( (-bHeight > dE)   || (bHeight < dE)    ) IsOut = 1; //Still in dE?

              // First track, then bin. This makes sure that the first entry is turn -1.
              // Turn 0 was handled in the forward tracking.  
              Phi -= dE*phiParam;
              dE  -= eV*(sin(Phi)-sPhiS); 
              //Turn -1 goes in part 0! Be Careful!
              DummyPhis[ -ThisTurn-1 ] = (int)floor( (Phi-PhiLoLimit)/phiBin ); 
              DummydEs[  -ThisTurn-1 ] = (int)floor( (dE+bHeight)/dEBin      ); 

            }//Array of bins for 1 particle each turn.

            //_______One Tracked_________________________

            //TimeDummy = clock();
            //Here the map uses turns > turns to store the negative turns information
            //So (1,2,3...Turns,-1,-2,-3...-Turns) This could alternatively be done
            // with a separate map for the -t turns. 
            if( !IsOut )
            {
              for( int TrnInd = Turns; TrnInd < 2*Turns; TrnInd += turnIncrement )
              {
                fiMap[ TrnInd ][ Pix ][ DummyPhis[ TrnInd % Turns ] ] += 1;  
              }
            }
          }//End BackwardTracking

          //TimeFill += clock() - TimeDummy;
        }//subj
      }//subi    
    }//binj
    #ifdef _LOUD_MAPS
      std::cout << "Row "<< bini <<" of "<< iSize << " tracked." << std::endl;
    #endif
  }//bini

  end = clock();
  #ifdef _STANDALONE_VERBOSE
    std::cout << "Maps Created." << std::endl;
  #endif 
  //std::cout << "Tracking complete: "<< (float)(end-start)/CLOCKS_PER_SEC << " sec." << std::endl;
  //std::cout << "\tTracking: "<< (float)(TimeTrack)/CLOCKS_PER_SEC << " sec." << std::endl;
  //std::cout << "\tFilling:  "<< (float)(TimeFill)/CLOCKS_PER_SEC << " sec." << std::endl;

  if( shouldWriteToFile )
  { 
    writeMapFiles( fiMap, kIGOR );
    logMap(); //_Update Log 
    //function for reading should fillnSubJals() else it won't be filled!
  }else{
    mapToLocalVectors( fiMap, kIGOR );
    fillNSubJ();
  }

  #ifdef _LOUD_MAPS
    std::cout << "mapLength = " << mapLength << std::endl;

    std::cout << "PixV.size() = " << PixV.size() << std::endl;
    std::cout << "BinV.size() = " << BinV.size() << std::endl;
    std::cout << "ValV.size() = " << ValV.size() << std::endl;
  #endif
};//End igorsMaps
