#include "TomographyAlgorithm.h"
#include <iostream>
#include <vector>
#include <math.h>
//Terminology comes from R.Gordon, A Tutorial On ART, IEEE Trans. Nucl. Sci. Vol. NS-21, June 1974
//It's at R.Gordon's Univ. Manitoba website. You should read it. 

void fullyConstrainedART2( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//This is a fully constrained ART algorithm.
	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV  = (int *)&(maps->PixV)[0];
	const int *BinV  = (int *)&(maps->BinV)[0];
	const int *ValV  = (int *)&(maps->ValV)[0];
	const int *nSubs = (int *)&(maps->nSubJ)[0];
	const int *BinT  = (int *)&(maps->BinTot)[0];
	const int *PixT  = (int *)&(maps->PixTot)[0];

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);

	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin ) NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
		//Normalization += frameTotal;
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); 
	}
	//Normalization /= numberFrames; //Normalization is to mean charge/frame in input data.
	
	int   ConvergenceFlag = 0;
	float Error;

	std::vector<float> InterReco(TotalPixels  , 0); //Intermediate image used in ART 2
		
	double *DiscrepancyPtr    = (double *)&((*Discrepancy)[0]); //These are used instead of the vectors. I'll fix it and use iterators. 
	float  *NormalizedDataPtr = (float *)&(NormalizedData[0]);  //These are used instead of the vectors. I'll fix it and use iterators. 
	float  *InterRecoPtr      = (float *)&(InterReco[0]);       //These are used instead of the vectors. I'll fix it and use iterators. 

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Calculate Error Profiles + Discrepancy______________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		//convergence is calculated as fractional change in discrepancy from iteration to iteration.
		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float cSubJ, nSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm uses that approximation and c_j is calcualted as
			//the total contributing to binj / sqrtPPP. It doesn't work very well.

			int bin = *(BinV + map_entry);
			cSubJ = (float)( *(BinT  + bin ) ) / (float)sqrtPPCell; // ( Particles contributing to Bin j ) / ( SqrtPPP )
			*(InterRecoPtr + *(PixV + map_entry) ) += *(Projections + bin) * 1.0 / cSubJ;
		}
		//____________________________________________________________BP

		//Squelch < 0 && > Normalization ____________________________SQ
		//Here the intermediate image is squelched and assigned to actual image before forward projection.
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			 *(Image+pix) = *(InterRecoPtr+pix) < 0.0           ? 0.0           : *( InterRecoPtr + pix ); //Lower Limit
			 *(Image+pix) = *(Image+pix       ) > Normalization ? Normalization : *(Image+pix)           ; //Upper Limit
		}
		//____________________________________________________________SQ

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * ( *(ValV + map_entry) );
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int FrameStart = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix) *(Image+pix)          *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1.                
}

void fullyConstrainedART( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//This is a fully constrained ART algorithm.
	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->nSubJ)[0];

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);

	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal                        += (*DataToRecon)[ bin + FrameStart ]               ; //Calculate normalization constant
		for( int bin = 0; bin < ptsperprofile; ++bin ) NormalizedData[ bin + FrameStart ] = (*DataToRecon)[ bin + FrameStart ] * 1/frameTotal; //Normalize
		
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}
	
	
	int   ConvergenceFlag = 0;
	float Error;
		
	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{

		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if(iter >= 1 )ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float nSubJ, cSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP )
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm uses that approximation and c_j is calcualted as
			//the total contributing to binj / sqrtPPP. It doesn't work very well.

			cSubJ = (float)( *(BinT + *(BinV + map_entry) ) ) / (float)sqrtPPCell;
			*(Image + *(PixV + map_entry) ) += *(Projections + *(BinV + map_entry)) * 1.0 / cSubJ ;
		}
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{	
			*(Image + pix) = *(Image + pix) < 0.0           ? 0.0           : *(Image + pix);
			*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}

		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * (*(ValV + map_entry));
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix) *(Image+pix) *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1.  
}
	

void fullyConstrainedDampedART( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//This is a fully constrained damped ART algorithm.
	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	//This is an implentation of a fully constrained ART algorithm.
	//Image and profiles renormalzied to 'Normalization' before writing to 'Projections'
	//Do a single bunch here, then implement the parallization as parallel 
	//calls to this function. This will allow interaction with gSystem

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);

	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin ) NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}
	
	int   ConvergenceFlag = 0;
	float Error;
		
	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	float dampingNow;
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		dampingNow = iter == 0 ? 1 : dampingFactor;
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if(iter >= 1 )ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float nSubJ, cSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm uses that approximation and c_j is calcualted as
			//the total contributing to binj / sqrtPPP. It doesn't work very well.

			cSubJ = (float)( *(BinT + *(BinV + map_entry) ) ) / (float)sqrtPPCell;
			*(Image + *(PixV + map_entry) ) += dampingNow * *(Projections + *(BinV + map_entry)) * 1.0 / cSubJ;
		}
		//____________________________________________________________BP
		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(Image + pix) < 0.0           ? 0.0           : *(Image + pix);
			*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * (*(ValV + map_entry));
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1.  
};


void fullyConstrainedART3( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//Damping factor times the max input signal is taken as the error factor here. A bit of a kostyl, as they say.

	//This is a fully constrained damped ART algorithm. That allows errors in the input
	//profiles. A factor gamma is used to determine the correction to the image. 
	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	float inputMax = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);
	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin )
		{
			NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
			inputMax = NormalizedData[ bin + FrameStart ] > inputMax ? NormalizedData[ bin + FrameStart ] : inputMax;
		}
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}

	float errMax = dampingFactor * inputMax;
	//std::cout << "errMax = " << errMax << std::endl;
	int   ConvergenceFlag = 0;
	float Error;
		
	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float gamma, cSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm uses that approximation and c_j is calcualted as
			//the total contributing to binj / sqrtPPP. It doesn't work very well.

			// Whoa, now! What you doin! 
			// For big gamma take gamma. If within error, leave it. If within twice error correct by error in the right direction.
			// gamma = ( fabs(gamma)  > 2*errMax ) ? gamma : ( fabs(gamma) < errMax ? 0.0 : ( gamma < 0.0 ?  2*( gamma + errMax ) : 2*( gamma - errMax ) ) );
			gamma = *(Projections + *(BinV + map_entry));

			if( fabs(gamma)  > 2*errMax )
			{
				gamma = gamma;
			}else if( fabs(gamma) < errMax ){
				gamma = 0.0;
			}else if( gamma < 0.0 ){
				gamma = 2*( gamma + errMax );
			}else if( gamma > 0.0 ){
				gamma = 2*( gamma - errMax );
			}

			cSubJ = (float)( *(BinT + *(BinV + map_entry) ) ) / (float)sqrtPPCell;
			*(Image + *(PixV + map_entry) ) += gamma * 1.0 / cSubJ ;
		}

		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(Image + pix) < 0.0           ? 0.0           : *(Image + pix);
			*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * (*(ValV + map_entry));
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1. 
};

void testART( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//This function is used for testing. Don't ever rely on it.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	//This is an implentation of a fully constrained ART algorithm.
	//Image and profiles renormalzied to 'Normalization' before writing to 'Projections'
	//Do a single bunch here, then implement the parallization as parallel 
	//calls to this function. This will allow interaction with gSystem

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	float inputMax = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);
	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin )
		{
			NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
			inputMax = NormalizedData[ bin + FrameStart ] > inputMax ? NormalizedData[ bin + FrameStart ] : inputMax;
		}
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}

	float errMax = dampingFactor * inputMax;
	//std::cout << "errMax = " << errMax << std::endl;
	int   ConvergenceFlag = 0;
	float Error;
		
	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float gamma, cSubJ, nSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//I need to consider the justification for this, but it seems to 
			//handle the separatrix MUCH better than the 'actual' ART without having to keep
			//track of what's going on there. The discrepancies even seem to be lower. Need to dedice what's happening here.
			cSubJ = (float)(*(BinT + *(BinV + map_entry) ) );
			gamma = *(Projections + *(BinV + map_entry));
			*(Image + *(PixV + map_entry) ) += gamma * *(ValV + map_entry) / cSubJ;
		}
		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(Image + pix) < 0.0           ? 0.0           : *(Image + pix);
			//*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			cSubJ = (*(ValV + map_entry));
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * cSubJ;
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1. 
	std::cout << "Finished running test algo." << std::endl; 
};

void probabilisticART( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//The backprojection step here does not follow Gordon, but it seems to work the best.
	//the backprojection weight is fi = pj * (Total from Cell i) / (Total contributing to Bin j)
	//This is how Duncan and I were thinking about the problem.

	//This is ALMOST a fully constrained ART algorithm.

	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	//This is an implentation of a fully constrained ART algorithm.
	//Image and profiles renormalzied to 'Normalization' before writing to 'Projections'
	//Do a single bunch here, then implement the parallization as parallel 
	//calls to this function. This will allow interaction with gSystem

	const int TotalBins   = ptsperprofile * numberFrames;
	
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	float inputMax = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);
	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin )
		{
			NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
			inputMax = NormalizedData[ bin + FrameStart ] > inputMax ? NormalizedData[ bin + FrameStart ] : inputMax;
		}
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}

	int   ConvergenceFlag = 0;
	float Error;
		
	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float delta, cSubJ, nSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm eschews that entirely. Here backprojection is done as
			// c_j = ( Total contribution to Bin j ) / ( contribution from pixel i )
			// since we divide by c_j this is effectively, 1/c_j is effectively the probability that
			// the charge in delta came from pixel i.

			//I need to consider the justification for this, but it seems to 
			//handle the separatrix MUCH better than the 'actual' ART without having to keep
			//track of what's going on there. The discrepancies even seem to be lower. Need to dedice what's happening here.
			cSubJ = (float)(*(BinT + *(BinV + map_entry) ) ) / (float)(*(ValV + map_entry));
			delta = *(Projections + *(BinV + map_entry));
			*(Image + *(PixV + map_entry) ) += delta / cSubJ;
		}
		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(Image + pix) < 0.0           ? 0.0           : *(Image + pix);
			//*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			cSubJ = (*(ValV + map_entry));
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * cSubJ;
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1. 
};

void probabilisticART2( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//The backprojection step here does not follow Gordon, but it seems to work the best.
	//the backprojection weight is fi = pj * (Total from Cell i) / (Total contributing to Bin j)
	//This is how Duncan and I were thinking about the problem.

	//This is ALMOST a fully constrained ART algorithm.

	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	const Tomography_Config tomoStruct = data->getTomography();

	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;
	const int sqrtPPCell         = tomoStruct.SqrtPPP;

	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);

	const int mapLength = maps->mapLength; 
	
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	//This is an implentation of a fully constrained ART algorithm.
	//Image and profiles renormalzied to 'Normalization' before writing to 'Projections'
	//Do a single bunch here, then implement the parallization as parallel 
	//calls to this function. This will allow interaction with gSystem

	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	float Normalization = 0.0;
	float inputMax = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);
	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin )
		{
			NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
			inputMax = NormalizedData[ bin + FrameStart ] > inputMax ? NormalizedData[ bin + FrameStart ] : inputMax;
		}
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}

	int   ConvergenceFlag = 0;
	float Error;

	std::vector<float> InterReco(TotalPixels  , 0); //Intermediate image used in ART 2

	double *DiscrepancyPtr   = (double *)&((*Discrepancy)[0]);
	float *NormalizedDataPtr = (float  *)&(NormalizedData[0]);
	float  *InterRecoPtr      = (float *)&(InterReco[0]);      

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		float delta, cSubJ, nSubJ;
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm eschews that entirely. Here backprojection is done as
			// c_j = ( Total contribution to Bin j ) / ( contribution from pixel i )
			// since we divide by c_j this is effectively, 1/c_j is effectively the probability that
			// the charge in delta came from pixel i.

			//I need to consider the justification for this, but it seems to 
			//handle the separatrix MUCH better than the 'actual' ART without having to keep
			//track of what's going on there. The discrepancies even seem to be lower. Need to dedice what's happening here.
			cSubJ = (float)(*(BinT + *(BinV + map_entry) ) ) / (float)(*(ValV + map_entry));
			delta = *(Projections + *(BinV + map_entry));
			*(InterRecoPtr + *(PixV + map_entry) ) += delta / cSubJ;
		}
		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(InterRecoPtr + pix) < 0.0 ? 0.0 : *(InterRecoPtr + pix);
			//*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			cSubJ = (*(ValV + map_entry));
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * cSubJ;
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm; //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1. 
};

void probabilisticART3( const int ptsperprofile, const int bunch, const MapClass* maps, dataObject* data )
{
	//The backprojection step here does not follow Gordon, but it seems to work the best.
	//the backprojection weight is fi = pj * (Total from Cell i) / (Total contributing to Bin j)
	//This is how Duncan and I were thinking about the problem.

	//This is ALMOST a fully constrained ART algorithm.

	//The constraint is very weak, charge in a single cell is not allowed to be larger
	//than the mean charge for an entire frame.
	//Image and profiles renormalzied to mean charge in a frame of input data calculate
	//here as 'Normalization' before the function returns.

	//Single bunch here allows interaction with gSystem for progress bar, etc. without
	//including any of the higher level program stuff.

	//Parameters
	const Tomography_Config tomoStruct = data->getTomography();
	const int numberFrames       = tomoStruct.Frames;
	const int MaxIterations      = tomoStruct.MaxIterations;
	const float ConvergenceLimit = tomoStruct.ConvergenceLimit;
	const float dampingFactor    = tomoStruct.dampingFactor;

	//Data
	const std::vector< float > *DataToRecon = data->getDataVec(bunch);
	std::vector<double>  *Discrepancy = data->getDiscVec(bunch);
	float *Image                      = data->getImage(bunch);
	float *Projections                = data->getProfile(bunch);
	
	//Map Stuff
	const int mapLength = maps->mapLength; 
	const int *PixV = (int *)&(maps->PixV)[0];
	const int *BinV = (int *)&(maps->BinV)[0];
	const int *ValV = (int *)&(maps->ValV)[0];
	const int *BinT = (int *)&(maps->BinTot)[0];

	//Derived parameters
	const int TotalBins   = ptsperprofile * numberFrames;
	const int pixelsPhi   = data->getImageWidth();
	const int pixelsdE    = data->getImageHeight();
	const int TotalPixels = pixelsPhi * pixelsdE;

	//Normalize the input data
	float inputMax = 0.0;
	float Normalization = 0.0;
	std::vector< float > NormalizedData(TotalBins, 0);
	for(int frame = 0; frame < numberFrames; ++frame)
	{
		float frameTotal = 0.0;
		int   FrameStart = frame*ptsperprofile;
		for( int bin = 0; bin < ptsperprofile; ++bin ) frameTotal += (*DataToRecon)[ bin + FrameStart ]; 
		for( int bin = 0; bin < ptsperprofile; ++bin )
		{
			NormalizedData[ bin + FrameStart ] = (*DataToRecon)[bin + FrameStart ] * 1/frameTotal; 
			inputMax = NormalizedData[ bin + FrameStart ] > inputMax ? NormalizedData[ bin + FrameStart ] : inputMax;
		}
		Normalization = Normalization + (frameTotal - Normalization)/(frame+1); //Calculating the mean as we go.
	}

	float errMax = dampingFactor * inputMax;
	float Error;
	int   ConvergenceFlag = 0;

	double *DiscrepancyPtr    = (double *)&((*Discrepancy)[0]);
	float  *NormalizedDataPtr = (float  *)&(NormalizedData[0]);

	///DI = DIsc. calc, B(F)P = Back(Forward) Project, NI(P) = Normal Image(Profiles)
	for(int iter = 0; (iter < MaxIterations)&&(ConvergenceFlag == 0); ++iter)
	{
		//Take difference_____________________________________________DI
		//Projections = Data - Projections;
		(*Discrepancy)[iter] = 0.0;
		for(int bin = 0; bin < TotalBins; ++bin)
		{
			Error                     = *(NormalizedDataPtr + bin) - *(Projections + bin); 
			*(Projections + bin)      = Error;
			*(DiscrepancyPtr + iter) += (float)Error*Error; //Gordon defines this with 1/Nj for each bin.
		}
		
		*(DiscrepancyPtr + iter) /= (float)TotalBins;
		*(DiscrepancyPtr + iter)  = sqrt( *(DiscrepancyPtr + iter) );

		if( iter >= 1 ) ConvergenceFlag = (fabs( *(DiscrepancyPtr + iter) - *(DiscrepancyPtr + iter-1) ) / (*(DiscrepancyPtr + iter-1)) <= ConvergenceLimit ? 1: 0);

		//____________________________________________________________DI

		//Backproject differences_____________________________________BP
		
		for(unsigned int map_entry = 0; map_entry < mapLength; ++map_entry)
		{	
			//In R.Gordon Tutorial on ART the backprojection step is given as:
			// f(q+1)_i = (f(q)_i + (p_j-p(q)_j)/(c_J*N_j) ) 
			// where p_j = data, p(q)_j = projection of the qth guess
			// and, N_j = total number of pixels contributing to bin p_j
			// and, c_j = (Area of jth ray) / (Area of all N_j pixels)
			// Assuming all cells have the same area, c_j = (Area of jth ray)/(N_j * Area of a Pixel)
			// This makes the c_j * N_j = (Area of the jth ray) / (Area of a pixel),
			// which in our case is: (Particles contributing to Bin j) / ( SqrtPPP ) (here contribution is from maps, not projections)
			// This does not handle the separatrix very well, since the approximation doesn't hold
			// and there are other problems there with particle motion. 

			//This version of the algorithm eschews that entirely. Here backprojection is done as
			// c_j = ( Total contribution to Bin j ) / ( contribution from pixel i )
			// since we divide by c_j this is effectively, 1/c_j is effectively the probability that
			// the charge in delta came from pixel i.

			float cSubJ = (float)(*(BinT + *(BinV + map_entry) ) ) / (float)(*(ValV + map_entry));
			float gamma = *(Projections + *(BinV + map_entry));

			if( fabs(gamma)  > 2*errMax )
			{
				gamma = gamma;
			}else if( fabs(gamma) < errMax ){
				gamma = 0.0;
			}else if( gamma < 0.0 ){
				gamma = 2*( gamma + errMax );
			}else if( gamma > 0.0 ){
				gamma = 2*( gamma - errMax );
			}

			*(Image + *(PixV + map_entry) ) += gamma / cSubJ ;
		}
		
		//____________________________________________________________BP

		//Squelch < 0__________________________________NI
		for(int pix = 0; pix < TotalPixels; ++pix)
		{
			*(Image + pix) = *(Image + pix) < 0.0 ? 0.0 : *(Image + pix);
			//*(Image + pix) = *(Image + pix) > Normalization ? Normalization : *(Image + pix);
		}
		//____________________________________________________________NI

		//Forward project image_______________________________________FP
		for( int bin = 0; bin < TotalBins; ++bin ) *(Projections + bin) = 0; //Reset projections
		
		for( unsigned int map_entry = 0; map_entry < mapLength; ++map_entry )
		{
			*(Projections + *(BinV + map_entry) ) += *(Image + *(PixV + map_entry) ) * (*(ValV + map_entry));
		}
		
		//____________________________________________________________FP

		//Normalize Projections_______________________________________NP
		for( int frame = 0; frame < numberFrames; ++frame )
		{
			float ProfileNorm = 0;
			int   FrameStart  = frame * ptsperprofile;

			for( int bin = 0; bin < ptsperprofile; ++bin ) ProfileNorm += *(Projections + FrameStart + bin );
			
			ProfileNorm = ProfileNorm > 0 ? 1/ProfileNorm : 0;

			for( int bin = 0; bin < ptsperprofile; ++bin ) *(Projections + FrameStart + bin ) *= ProfileNorm;
		}
		//____________________________________________________________NP
		#ifdef _LOUD_RECONSTRUCTIONS
			std::cout << "Iteration " << iter << " complete." << std::endl;
		#endif
	}//End Iteration

	//Final normalization to 'Normalization'
	float ImageNorm = 0;
	for(int pix = 0; pix < TotalPixels; ++pix) ImageNorm += *(Image+pix);
	ImageNorm = (ImageNorm > 0 ? Normalization/ImageNorm : 0);
	for(int pix = 0; pix < TotalPixels; ++pix ) *(Image+pix)         *= ImageNorm;     //Images need to be completely normalized. 
	for(int bin = 0; bin < TotalBins  ; ++bin ) *(Projections + bin) *= Normalization; //Projections were previously normalized to 1. 
};