#include "dataObject.h"

dataObject::dataObject( const Machine_Config machine, const Tomography_Config tomography )
{
	clearHeader();
	setMachine(machine);
	setTomography(tomography);
}

dataObject::dataObject( const complete_data_header_type Config )
{
	//Constuctor only sets the Tomography and Machine parameters.
	clearHeader(); 
	setMachine( Config.machine );
	setTomography( Config.tomography );
}

dataObject::~dataObject()
{
	std::vector< std::vector< float > >::iterator it;
	for(it = InputProfiles.begin() ; it != InputProfiles.end() ; ++it) (*it).resize(0);
	for(it = OutputProfiles.begin(); it != OutputProfiles.end(); ++it) (*it).resize(0);
	for(it = OutputImages.begin()  ; it != OutputImages.end()  ; ++it) (*it).resize(0);
}

void dataObject::padData( int halfPad )
{
	//Pads data equally on each side. Also updates ptsperprofile in the header.
	int fullPad   = 2 * halfPad;
	int oldLength = objectHeader.dataAcquisition.ptsperprofile;
	objectHeader.dataAcquisition.ptsperprofile += fullPad;

	std::vector< std::vector<float> >::iterator bunchIt    = InputProfiles.begin();
	std::vector< std::vector<float> >::iterator profilesIt = OutputProfiles.begin();
	std::vector< std::vector<float> >::iterator imagesIt   = OutputImages.begin();

	for( bunchIt; bunchIt != InputProfiles.end(); ++bunchIt )
	{
		//Insert all the inner paddings
		for( int frame = 1; frame < objectHeader.dataAcquisition.num_frames; ++frame )
		{
			(*bunchIt).insert((*bunchIt).begin() + frame * oldLength + (frame - 1) * fullPad, fullPad, 0);
		}
		
		//Pad ends
		(*bunchIt).insert((*bunchIt).begin(), halfPad, 0);
		(*bunchIt).insert((*bunchIt).end()  , halfPad, 0);
	}
}

void dataObject::clear()
{
	clear_profile_data_header_type( objectHeader.dataAcquisition );
	InputProfiles.clear();
	OutputProfiles.clear();
	OutputImages.clear();
}

bool dataObject::readInPRPFile( const char* filename )
{
	clear();

	bool didRead = false;

	FILE* temp = fopen( filename, "rb" );

	fseek( temp, 0, SEEK_END );
	long realSize = ftell( temp );
	fseek( temp, 0, SEEK_SET );

	if( realSize >= sizeof( profile_data_header_type ) ) // make sure it's at least as big as a header
	{
		//Don't read in a struct this way unless it's packed!
		fread( (char *)&objectHeader.dataAcquisition, sizeof(profile_data_header_type), 1, temp );
		long calcSize = objectHeader.dataAcquisition.ptsperprofile * objectHeader.dataAcquisition.num_frames * objectHeader.dataAcquisition.num_bunches * sizeof(float) + sizeof(profile_data_header_type);
		if( realSize == calcSize )
		{
			//Read entire bunch into object. 
			long NPointsTot = (objectHeader.dataAcquisition.ptsperprofile)*(objectHeader.dataAcquisition.num_frames); //Length of one bunch*DataFrames
			std::vector<float> Dummy(NPointsTot);

			//This might not work like I think it will need to test it. 
			hardPrep();
			for(int bunch = 0; bunch < objectHeader.dataAcquisition.num_bunches; ++bunch)
			{

				fread( &Dummy[0], sizeof(float), NPointsTot, temp);
				//fread( &InputProfiles[bunch][0], sizeof(float), NPointsTot, temp);
				//Do this with vector iterators. 
				for(int bin = 0; bin < NPointsTot; ++bin)
				{
					InputProfiles[bunch][ bin ] = (float)Dummy[ bin ]; //Maybe include a frames offset here. See function in TomographyConfig.h
				} 
			}
			didRead = true;
		}
	}
	#ifdef _STANDALONE_VERBOSE
		if(didRead == false) std::cout << "File format incorrect or file is corrupt."  << std::endl;
	#endif

	fclose(temp);
	return didRead;	
}

void dataObject::readIn2DVector( const std::vector< std::vector< float > > input, const profile_data_header_type head )
{
	setHeader( head );
	InputProfiles.resize( input.size() );
	for(int b = 0; b < input.size(); ++b ) InputProfiles[b] = input[b];
}

void dataObject::bunchFromArray( const float *array, const int size, const int bunch )
{
	//remember, if bunch = 0, size = 1
	if( InputProfiles.size() < bunch + 1 ) InputProfiles.resize( bunch + 1 );
	//if( InputProfiles[bunch].size() < size ) InputProfiles[bunch].resize( size, 0.0 );
	InputProfiles[bunch].clear();
	for(int i = 0; i < size; ++i) (InputProfiles[bunch]).push_back( (*(array + i)) );

}

void dataObject::bunchFromArray( const float *array, const profile_data_header_type* head )
{
	//This just calls the above function for every bunch. 
	setHeader( (*head) );
	hardPrep();

	int bunchSize = head->ptsperprofile * head->num_frames;

	for(int bunch = 0; bunch < head->num_bunches; ++bunch) bunchFromArray( array + bunch * bunchSize, bunchSize, bunch );
}


void dataObject::reconstructProfiles( const MapClass* maps, const int algo )
{

	int progress = 0;
	prepForTomography();

	#pragma omp parallel for
	for(int bunch = 0; bunch < objectHeader.dataAcquisition.num_bunches; ++bunch)
	{	
		this->reconstructSingle(maps, bunch, algo);
		
		#ifdef _STANDALONE_VERBOSE
			++progress; //If you don't have a counter openmp splits the loop and bunch updates seemingly at random.
			std::cout << "Bunch "<< progress << "/" << objectHeader.dataAcquisition.num_bunches << "complete" << std::endl;
		#endif
	}
}

void dataObject::reconstructSingle( const MapClass* maps, const int bunch, const int algo )
{
	//Don't want to prep inside this one because this function gets threaded.
	if(bunch < objectHeader.dataAcquisition.num_bunches)
	{
		const int ptsper = objectHeader.dataAcquisition.ptsperprofile;
		switch(algo)
		{
    		case kDFCART_ID:{
    			fullyConstrainedDampedART( ptsper, bunch, maps, this ); 
    		}
    		break;
			case kFCART_ID:{
				fullyConstrainedART( ptsper, bunch, maps, this );
			}
			break;
			case kFCART2_ID:{
					fullyConstrainedART2( ptsper, bunch, maps, this );
			}
			break;
			case kFCART3_ID:{
				fullyConstrainedART3( ptsper, bunch, maps, this );
			}
			break;
			case kTESTING_ID:{	
				testART( ptsper, bunch, maps, this );
			}
			break;
			case kPROBART_ID:{
				probabilisticART( ptsper, bunch, maps, this );
			}
			break;
			case kPROBART2_ID:{
				probabilisticART2( ptsper, bunch, maps, this );
			}
			break;
			case kPROBART3_ID:{
				probabilisticART3( ptsper, bunch, maps, this );
			}
			break;
		}	        	
	}
}

void dataObject::writeImages( const char* file )
{
	std::ofstream outfile;
	outfile.open( file, std::ios::out|std::ios::binary );

	int pixelsPhi = objectHeader.tomography.pixelsPhi;
	int pixelsdE  = objectHeader.tomography.pixelsdE;

	outfile.write( (char *)&pixelsPhi, sizeof(int) );
	outfile.write( (char *)&pixelsdE , sizeof(int) );

	int bunches = OutputProfiles.size();
	outfile.write( (char *)&bunches, sizeof(int) );

	for(int bunch = 0; bunch < OutputImages.size(); ++bunch)
	{ 
		outfile.write( (char *)&OutputImages[bunch][0], sizeof(float)*OutputImages[bunch].size() );
	}

	outfile.close();
}

void dataObject::writeProfiles( const char* file )
{
	std::ofstream outfile;
	outfile.open( file, std::ios::out|std::ios::binary );

	int ptsperprofile = objectHeader.dataAcquisition.ptsperprofile; 
	outfile.write( (char *)&ptsperprofile, sizeof(int) );

	int bunches = OutputProfiles.size();
	outfile.write( (char *)&bunches, sizeof(int) );

	for(int bunch = 0; bunch < OutputProfiles.size(); ++bunch)
	{ 
		outfile.write( (char *)&OutputProfiles[bunch][0], sizeof(float)*OutputProfiles[bunch].size() );
	}

	outfile.close();		
}

void dataObject::writeDiscrepancies( const char* file )
{
	std::ofstream outfile;
	outfile.open( file, std::ios::out|std::ios::binary );

	int size;

	int bunches = OutputProfiles.size();
	outfile.write( (char *)&bunches, sizeof(int) );

	for(int bunch = 0; bunch < bunches; ++bunch)
	{ 	
		size = Discrepancies[bunch].size();
		outfile.write( (char *)&size, sizeof(int) );
		outfile.write( (char *)&Discrepancies[bunch][0], sizeof(double)*Discrepancies[bunch].size() );
	}

	outfile.close();		
}

void dataObject::writeReconData( const char* file )
{
	std::ofstream outfile;
	outfile.open( file, std::ios::out|std::ios::binary );
	int pointsToWritePerBunch = objectHeader.dataAcquisition.ptsperprofile * objectHeader.tomography.Frames;
	for(int bunch = 0; bunch < InputProfiles.size(); ++bunch)
	{ 
		outfile.write( (char *)&InputProfiles[bunch][0], sizeof(float)*pointsToWritePerBunch );
	}

	outfile.close();		
}

void dataObject::writeFullData( const char* file )
{
	std::ofstream outfile;
	outfile.open( file, std::ios::out|std::ios::binary );

	for(int bunch = 0; bunch < InputProfiles.size(); ++bunch)
	{ 
		outfile.write( (char *)&InputProfiles[bunch][0], sizeof(float)*InputProfiles[bunch].size() );
	}

	outfile.close();		
}

void dataObject::printInputProfile( const int bunch )
{
	if(bunch <= InputProfiles.size())
	{
		for(int bin = 0; bin < InputProfiles[bunch].size(); ++bin)
		{
			std::cout << "Bunch "<< bunch << " Input[" << bin << "] = " << InputProfiles[bunch][bin] << std::endl;
		}
	}
}

void dataObject::printHeader(){          print_complete_data_header_type( objectHeader );                }
void dataObject::printDataAcquisition(){ print_profile_data_header_type( objectHeader.dataAcquisition ); }
void dataObject::printTomography(){      print_Tomography_Config( objectHeader.tomography );             }
void dataObject::printMachine(){         print_Machine_Config( objectHeader.machine );                   }

void dataObject::clearHeader()
{
	clear_profile_data_header_type( objectHeader.dataAcquisition );
}

complete_data_header_type dataObject::getConfig()
{
	return objectHeader;
}

profile_data_header_type dataObject::getDataAcquisition()
{
	return objectHeader.dataAcquisition;
}

Tomography_Config dataObject::getTomography()
{
	return objectHeader.tomography;
}

Machine_Config dataObject::getMachine()
{
	return objectHeader.machine;
}

float* dataObject::getData( const int bunch )
{
	float* returnValue = ( bunch < InputProfiles.size() ) ? (float *)&InputProfiles[bunch][0] : NULL;
	return returnValue;
}

float* dataObject::getImage( const int bunch )
{
	return bunch < OutputImages.size() ? (float *)&OutputImages[bunch][0] : NULL;
}

float* dataObject::getProfile( const int bunch )
{	
	return bunch < OutputProfiles.size() ? (float *)&OutputProfiles[bunch][0] : NULL;
}

double* dataObject::getDisc( const int bunch )
{
	return bunch < Discrepancies.size() ? (double *)&Discrepancies[bunch][0] : NULL;
}

std::vector<float>* dataObject::getDataVec(const int bunch)
{
	return bunch < InputProfiles.size() ? &InputProfiles[bunch] : &InputProfiles[InputProfiles.size()];
}

std::vector<float>* dataObject::getImageVec(const int bunch)
{
	return bunch < OutputImages.size() ? &OutputImages[bunch] : &OutputImages[OutputImages.size()];
}

std::vector<float>* dataObject::getProfileVec(const int bunch)
{
	return bunch < OutputProfiles.size() ? &OutputProfiles[bunch] : &OutputProfiles[OutputImages.size()];
}

std::vector<double>* dataObject::getDiscVec(const int bunch)
{
	return bunch < Discrepancies.size() ? &Discrepancies[bunch] : &Discrepancies[Discrepancies.size()];
}

void dataObject::setHeader( const profile_data_header_type head )
{
	copyProfile_data_header_type( &head, &(objectHeader.dataAcquisition) );
	//hardPrep(); //These are dangerous. If there is junk in a struct get bad alloc's
}

void dataObject::setMachine( const Machine_Config machine )
{
  	copyMachine( &machine, &(objectHeader.machine) );
  	//hardPrep(); //These are dangerous. If there is junk in a struct get bad alloc's
}

void dataObject::setTomography( const Tomography_Config tomography )
{
	copyTomography( &tomography, &(objectHeader.tomography) );
	//hardPrep(); //These are dangerous. If there is junk in a struct get bad alloc's
}

void dataObject::makeConsistent()
{
	if( objectHeader.tomography.Frames > objectHeader.dataAcquisition.num_frames ) objectHeader.tomography.Frames = objectHeader.dataAcquisition.num_frames;
}

//This should ONLY be used at initilization time when you're sure you've got a dataAcquisition struct
//garbage in a struct will mean you try to initialize vectors that are too large.
void dataObject::hardPrep()
{	
	if( objectHeader.tomography.Frames > objectHeader.dataAcquisition.num_frames ) objectHeader.tomography.Frames = objectHeader.dataAcquisition.num_frames;
	int num_bunches = objectHeader.dataAcquisition.num_bunches;
	int ptsperprofile = objectHeader.dataAcquisition.ptsperprofile;
	
	int pixelsPhi = objectHeader.tomography.pixelsPhi; 
	int pixelsdE  = objectHeader.tomography.pixelsdE ; 

	pixelsPhi = objectHeader.tomography.pixelsPhi  <= 0 ? ptsperprofile : pixelsPhi;
	pixelsdE  = objectHeader.tomography.pixelsdE   <= 0 ? ptsperprofile : pixelsdE ;

	InputProfiles.clear();
	OutputProfiles.clear();
	OutputImages.clear();
	Discrepancies.clear();

	InputProfiles.resize(  num_bunches );
	OutputProfiles.resize( num_bunches );
	OutputImages.resize(   num_bunches );
	Discrepancies.resize(  num_bunches );

	int totalPixels = pixelsPhi * pixelsdE;
	int totalBinsData  = objectHeader.dataAcquisition.num_frames * objectHeader.dataAcquisition.ptsperprofile;
	int totalBinsRecon = objectHeader.tomography.Frames * ptsperprofile;

	for(int bunch = 0; bunch < num_bunches; ++bunch)
	{
		InputProfiles[bunch].resize(  totalBinsData  , 0.0);
		OutputProfiles[bunch].resize( totalBinsRecon , 0.0);
		OutputImages[bunch].resize(   totalPixels    , 0.0);
		Discrepancies[bunch].resize(  objectHeader.tomography.MaxIterations, -1.0);
	}
}

void dataObject::prepForTomography()
{	
	//This is so not threadsafe, just leave your threads at home if you call this bad mutha. 
	//Similar to hard prep, but doesn't prep input data.
	if( objectHeader.tomography.Frames > objectHeader.dataAcquisition.num_frames ) objectHeader.tomography.Frames = objectHeader.dataAcquisition.num_frames;
	int num_bunches = objectHeader.dataAcquisition.num_bunches;
	int ptsperprofile = objectHeader.dataAcquisition.ptsperprofile;

	int pixelsPhi = objectHeader.tomography.pixelsPhi; 
	int pixelsdE  = objectHeader.tomography.pixelsdE ; 

	pixelsPhi = objectHeader.tomography.pixelsPhi  <= 0 ? ptsperprofile : pixelsPhi;
	pixelsdE  = objectHeader.tomography.pixelsdE   <= 0 ? ptsperprofile : pixelsdE ;

	//Clear first! if vector.size() = x and we do vector.resize(x, 0) NOTHING HAPPENS!
	OutputProfiles.clear();
	OutputImages.clear();
	Discrepancies.clear();

	OutputProfiles.resize( num_bunches );
	OutputImages.resize(   num_bunches );
	Discrepancies.resize(  num_bunches );

	int totalPixels = pixelsPhi * pixelsdE;
	int totalBinsRecon = objectHeader.tomography.Frames * objectHeader.dataAcquisition.ptsperprofile;
	
	//Use for arbitrary image size.	
	//totalPixels = imagedESize * imagePhiSize;

	for(int bunch = 0; bunch < num_bunches; ++bunch)
	{
		OutputProfiles[bunch].resize( totalBinsRecon                        , 0.0);
		OutputImages[bunch].resize(   totalPixels                           , 0.0);
		Discrepancies[bunch].resize(  objectHeader.tomography.MaxIterations ,-1.0);
	}

}

void dataObject::killFrame(int frame)
{	
	std::vector<float>::iterator kill_It;
	int ptsperprofile = objectHeader.dataAcquisition.ptsperprofile;
	for(int bunch = 0; bunch < objectHeader.dataAcquisition.num_bunches; ++bunch)
	{
		InputProfiles[bunch].erase( InputProfiles[bunch].begin() + frame, InputProfiles[bunch].begin() + frame + ptsperprofile );
	}
	objectHeader.dataAcquisition.num_frames -= 1;
}

bool dataObject::setFramesToBest()
{
	bool success  = false;
	double halfPeriod = SynchrotronPeriod_Turns( objectHeader.machine ) / 2.0 / objectHeader.dataAcquisition.turns_per_trig;

	//It's not super cricial that this round perfectly.
	int integerHalfPeriod = (int)( halfPeriod + 0.5 );

	if( integerHalfPeriod < 0)
	{
		success = false;
		#ifdef _STANDALONE_VERBOSE
			std::cout << " Frames not set by setFramesToBest(). "<< std::endl;
			std::cout << " Synchrotron period calculated was less than 0." << std::endl;
		#endif

	}else if( integerHalfPeriod <= objectHeader.dataAcquisition.num_frames ){

		objectHeader.tomography.Frames = integerHalfPeriod;
		success = true;
		#ifdef _STANDALONE_VERBOSE
			std::cout << " Frames set to 1/2 synchrtron period = " << objectHeader.tomography.Frames << std::endl;
		#endif
	}else{
		objectHeader.tomography.Frames = objectHeader.dataAcquisition.num_frames;
		success = true;
		#ifdef _STANDALONE_VERBOSE
			std::cout << " Frames set to max number in data = " << objectHeader.tomography.Frames << std::endl;
		#endif
	}


	return success;

}

